#!/usr/bin/env python
# License: GPLv3 Copyright: 2010, Kovid Goyal <kovid@kovidgoyal.net>

import os
import socket
import textwrap
import time
from collections import defaultdict
from functools import partial
from itertools import repeat
from threading import Thread
from typing import TYPE_CHECKING

from qt.core import QDialog, QDialogButtonBox, QGridLayout, QIcon, QLabel, QLineEdit, QListWidget, QListWidgetItem, QPushButton, Qt

from calibre.constants import preferred_encoding
from calibre.customize.ui import available_input_formats, available_output_formats
from calibre.ebooks.metadata import authors_to_string
from calibre.gui2 import Dispatcher, config, error_dialog, gprefs, warning_dialog
from calibre.gui2.threaded_jobs import ThreadedJob
from calibre.library.save_to_disk import get_components
from calibre.utils.config import prefs, tweaks
from calibre.utils.icu import primary_sort_key
from calibre.utils.localization import _
from calibre.utils.resources import get_image_path as I
from calibre.utils.smtp import compose_mail, extract_email_address, sendmail
from calibre.utils.smtp import config as email_config
from polyglot.binary import from_hex_unicode

if TYPE_CHECKING:
    from calibre.gui2.ui import Main


class Worker(Thread):
    def __init__(self, func, args):
        super().__init__(name='EmailWorker', daemon=True, target=self.do_work, args=(func, args))
        self.exception = self.tb = None

    def do_work(self, func, args):
        # time.sleep(1000)
        try:
            func(*args)
        except Exception as e:
            import traceback

            self.exception = e
            self.tb = traceback.format_exc()


class Sendmail:
    MAX_RETRIES = 1
    TIMEOUT = 25 * 60  # seconds

    def __init__(self):
        self.calculate_rate_limit()
        self.last_send_time = time.time() - self.rate_limit

    def calculate_rate_limit(self):
        self.rate_limit = 1
        opts = email_config().parse()
        rh = opts.relay_host
        if rh:
            for suffix in tweaks['public_smtp_relay_host_suffixes']:
                if rh.lower().endswith(suffix):
                    self.rate_limit = tweaks['public_smtp_relay_delay']
                    break

    def __call__(self, attachment, aname, to, subject, text, log=None, abort=None, notifications=None):

        try_count = 0
        while True:
            if try_count > 0 and log is not None:
                log(f'\nRetrying in {self.rate_limit} seconds...\n')
            worker = Worker(self.sendmail, (attachment, aname, to, subject, text, log))
            worker.start()
            start_time = time.time()
            while worker.is_alive():
                worker.join(0.2)
                assert abort is not None
                if abort.is_set():
                    if log:
                        log('Sending aborted by user')
                    return
                if time.time() - start_time > self.TIMEOUT:
                    if log:
                        log('Sending timed out')
                    raise Exception(f'Sending email {subject!r} to {to!r} timed out, aborting')
            if worker.exception is None:
                if log:
                    log('Email successfully sent')
                return
            assert log is not None
            log.error('\nSending failed...\n')
            log.debug(worker.tb)
            try_count += 1
            if try_count > self.MAX_RETRIES:
                raise worker.exception

    def sendmail(self, attachment, aname, to, subject, text, log):
        logged = False
        while time.time() - self.last_send_time <= self.rate_limit:
            if not logged and self.rate_limit > 0:
                log(f'Waiting {self.rate_limit} seconds before sending, to avoid being marked as spam.\nYou can control this delay via Preferences->Tweaks')
                logged = True
            time.sleep(1)
        try:
            opts = email_config().parse()
            from_ = opts.from_
            if not from_:
                from_ = 'calibre <calibre@' + socket.getfqdn() + '>'
            with open(attachment, 'rb') as f:
                msg = compose_mail(from_, to, text, subject, f, aname)
            efrom = extract_email_address(from_)
            eto = []
            for x in to.split(','):
                eto.append(extract_email_address(x.strip()))

            def safe_debug(*args, **kwargs):
                try:
                    return log.debug(*args, **kwargs)
                except Exception:
                    pass

            relay = opts.relay_host
            if relay and relay == 'smtp.live.com':
                # Microsoft changed the SMTP server
                relay = 'smtp-mail.outlook.com'

            sendmail(
                msg,
                efrom,
                eto,
                localhost=None,
                verbose=1,
                relay=relay,
                username=opts.relay_username,
                password=from_hex_unicode(opts.relay_password),
                port=opts.relay_port,
                encryption=opts.encryption,
                debug_output=safe_debug,
            )
        finally:
            self.last_send_time = time.time()


gui_sendmail = Sendmail()


def is_for_kindle(to):
    return isinstance(to, str) and ('@kindle.com' in to or '@kindle.cn' in to or '@free.kindle.com' in to or '@free.kindle.cn' in to)


def send_mails(jobnames, callback, attachments, to_s, subjects, texts, attachment_names, job_manager):
    for name, attachment, to, subject, text, aname in zip(jobnames, attachments, to_s, subjects, texts, attachment_names):
        description = _('Email %(name)s to %(to)s') % dict(name=name, to=to)
        if isinstance(to, str) and (is_for_kindle(to) or '@pbsync.com' in to):
            # The PocketBook service is a total joke. It can't handle
            # non-ascii, filenames that are long enough to be split up, commas, and
            # the good lord alone knows what else. So use a random filename
            # containing only 22 English letters and numbers
            #
            # And since this email is only going to be processed by automated
            # services, make the subject+text random too as at least the amazon
            # service can't handle non-ascii text. I don't know what baboons
            # these companies employ to write their code. It's the height of
            # irony that they are called "tech" companies.
            # https://bugs.launchpad.net/calibre/+bug/1989282
            if not is_for_kindle(to):
                from calibre.utils.short_uuid import uuid4

                # Amazon nowadays reads metadata from attachment filename instead of
                # file internal metadata so don't nuke the filename.
                # https://www.mobileread.com/forums/showthread.php?t=349290
                aname = f'{uuid4()}.' + aname.rpartition('.')[-1]
            from calibre.utils.random_ua import random_english_text

            subject = random_english_text(min_words_per_sentence=3, max_words_per_sentence=9, max_num_sentences=1).rstrip('.')
            text = random_english_text()
        job = ThreadedJob('email', description, gui_sendmail, (attachment, aname, to, subject, text), {}, callback)
        job_manager.run_threaded_job(job)


def email_news(mi, remove, get_fmts, done, job_manager):
    opts = email_config().parse()
    accounts = [(account, [x.strip().lower() for x in x[0].split(',')]) for account, x in opts.accounts.items() if x[1]]
    sent_mails = []
    for i, x in enumerate(accounts):
        account, fmts = x
        files = get_fmts(fmts)
        files = [f for f in files if f is not None]
        if not files:
            continue
        if opts.tags.get(account, False) and not ({t.strip() for t in opts.tags[account].split(',')} & set(mi.tags)):
            continue
        attachment = files[0]
        to_s = [account]
        subjects = [_('News:') + ' ' + mi.title]
        texts = [_('Attached is the %s periodical downloaded by calibre.') % (mi.title,)]
        attachment_names = [mi.title + os.path.splitext(attachment)[1]]
        attachments = [attachment]
        jobnames = [mi.title]
        do_remove = []
        if i == len(accounts) - 1:
            do_remove = remove
        send_mails(
            jobnames,
            Dispatcher(partial(done, remove=do_remove)),
            attachments,
            to_s,
            subjects,
            texts,
            attachment_names,
            job_manager,
        )
        sent_mails.append(to_s[0])
    return sent_mails


plugboard_email_value = 'email'
plugboard_email_formats = ['epub', 'mobi', 'azw3']


class SelectRecipients(QDialog):  # {{{
    def __init__(self, parent=None):
        QDialog.__init__(self, parent)
        self._layout = l = QGridLayout(self)
        self.setLayout(l)
        self.setWindowIcon(QIcon(I('mail.png')))
        self.setWindowTitle(_('Select recipients'))
        self.recipients = r = QListWidget(self)
        l.addWidget(r, 0, 0, 1, -1)
        self.la = la = QLabel(_('Add a new recipient:'))
        la.setStyleSheet('QLabel { font-weight: bold }')
        l.addWidget(la, l.rowCount(), 0, 1, -1)

        self.labels = tuple(map(QLabel, (_('&Address'), _('A&lias'), _('&Formats'), _('&Subject'))))
        tooltips = (
            _('The email address of the recipient'),
            _('The optional alias (simple name) of the recipient'),
            _('Formats to email. The first matching one will be sent (comma separated list)'),
            _('The optional subject for email sent to this recipient'),
        )

        i = -1

        def do() -> QLineEdit:
            nonlocal i
            i += 1
            c = i % 2
            row = l.rowCount() - c
            self.labels[i].setText(str(self.labels[i].text()) + ':')
            l.addWidget(self.labels[i], row, (2 * c))
            le = QLineEdit(self)
            le.setToolTip(tooltips[i])
            self.labels[i].setBuddy(le)
            l.addWidget(le, row, (2 * c) + 1)
            return le

        self.address, self.alias = do(), do()
        self.formats, self.subject = do(), do()
        self.formats.setText(prefs['output_format'].upper())
        self.add_button = b = QPushButton(QIcon(I('plus.png')), _('&Add recipient'), self)
        b.clicked.connect(self.add_recipient)
        l.addWidget(b, l.rowCount(), 0, 1, -1)

        self.bb = bb = QDialogButtonBox(QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel)
        l.addWidget(bb, l.rowCount(), 0, 1, -1)
        bb.accepted.connect(self.accept)
        bb.rejected.connect(self.reject)
        self.setMinimumWidth(500)
        self.setMinimumHeight(400)
        self.resize(self.sizeHint())
        self.init_list()

    def add_recipient(self):
        to = str(self.address.text()).strip()
        if not to:
            return error_dialog(self, _('Need address'), _('You must specify an address'), show=True)
        from email.utils import parseaddr

        if not parseaddr(to)[-1] or '@' not in to:
            return error_dialog(self, _('Invalid email address'), _('The address {} is invalid').format(to), show=True)
        formats = ','.join([x.strip().upper() for x in str(self.formats.text()).strip().split(',') if x.strip()])
        if not formats:
            return error_dialog(self, _('Need formats'), _('You must specify at least one format to send'), show=True)
        opts = email_config().parse()
        if to in opts.accounts:
            return error_dialog(self, _('Already exists'), _('The recipient %s already exists') % to, show=True)
        acc = opts.accounts
        acc[to] = [formats, False, False]
        c = email_config()
        c.set('accounts', acc)
        alias = str(self.alias.text()).strip()
        if alias:
            opts.aliases[to] = alias
            c.set('aliases', opts.aliases)
        subject = str(self.subject.text()).strip()
        if subject:
            opts.subjects[to] = subject
            c.set('subjects', opts.subjects)
        self.create_item(alias or to, to, checked=True)

    def create_item(self, alias, key, checked=False):
        i = QListWidgetItem(alias, self.recipients)
        i.setFlags(Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsUserCheckable)
        i.setCheckState(Qt.CheckState.Checked if checked else Qt.CheckState.Unchecked)
        i.setData(Qt.ItemDataRole.UserRole, key)
        self.items.append(i)

    def init_list(self):
        opts = email_config().parse()
        self.items = []

        def sk(account):
            return primary_sort_key(opts.aliases.get(account) or account)

        for key in sorted(opts.accounts or (), key=sk):
            self.create_item(opts.aliases.get(key, key), key)

    def accept(self):
        if not self.ans:
            return error_dialog(self, _('No recipients'), _('You must select at least one recipient'), show=True)
        QDialog.accept(self)

    @property
    def ans(self):
        opts = email_config().parse()
        ans = []
        for i in self.items:
            if i.checkState() == Qt.CheckState.Checked:
                to = str(i.data(Qt.ItemDataRole.UserRole) or '')
                fmts = tuple(x.strip().upper() for x in (opts.accounts[to][0] or '').split(','))
                subject = opts.subjects.get(to, '')
                ans.append((to, fmts, subject))
        return ans


def select_recipients(parent=None):
    d = SelectRecipients(parent)
    if d.exec() == QDialog.DialogCode.Accepted:
        return d.ans
    return ()


# }}}


class EmailMixin:  # {{{
    def send_multiple_by_mail(self: Main, recipients, delete_from_library):
        sm = self.library_view.selectionModel()
        assert sm is not None
        ids = {self.library_view._model.id(r) for r in sm.selectedRows()}
        if not ids:
            return
        db = self.current_db
        db_fmt_map = {book_id: set((db.formats(book_id, index_is_id=True) or '').upper().split(',')) for book_id in ids}
        ofmts = {x.upper() for x in available_output_formats()}
        ifmts = {x.upper() for x in available_input_formats()}
        bad_recipients = {}
        auto_convert_map = defaultdict(list)

        for to, fmts, subject in recipients:
            rfmts = set(fmts)
            ok_ids = {book_id for book_id, bfmts in db_fmt_map.items() if bfmts.intersection(rfmts)}
            convert_ids = ids - ok_ids
            self.send_by_mail(to, fmts, delete_from_library, subject=subject, send_ids=ok_ids, do_auto_convert=False)
            if not rfmts.intersection(ofmts):
                bad_recipients[to] = (convert_ids, True)
                continue
            outfmt = tuple(f for f in fmts if f in ofmts)[0]
            ok_ids = {book_id for book_id in convert_ids if db_fmt_map[book_id].intersection(ifmts)}
            bad_ids = convert_ids - ok_ids
            if bad_ids:
                bad_recipients[to] = (bad_ids, False)
            if ok_ids:
                auto_convert_map[outfmt].append((to, subject, ok_ids))

        if auto_convert_map:
            titles = {book_id for x in auto_convert_map.values() for data in x for book_id in data[2]}
            titles = {db.new_api.field_for('title', book_id) for book_id in titles}
            if self.auto_convert_question(_('Auto convert the following books before sending via email?'), list(titles)):
                for ofmt, data in auto_convert_map.items():
                    ids = {bid for x in data for bid in x[2]}
                    data = [(to, subject) for to, subject, x in data]
                    self.iactions['Convert Books'].auto_convert_multiple_mail(ids, data, ofmt, delete_from_library)

        if bad_recipients:
            det_msg = []
            titles = {book_id for x in bad_recipients.values() for book_id in x[0]}
            titles = {book_id: db.new_api.field_for('title', book_id) for book_id in titles}
            for to, (ids, nooutput) in bad_recipients.items():
                msg = _('This recipient has no valid formats defined') if nooutput else _('These books have no suitable input formats for conversion')
                det_msg.append(f'{to} - {msg}')
                det_msg.extend('\t' + titles[bid] for bid in ids)
                det_msg.append('\n')
            warning_dialog(
                self,
                _('Could not send'),
                _('Could not send books to some recipients. Click "Show details" for more information'),
                det_msg='\n'.join(det_msg),
                show=True,
            )

    def send_by_mail(self: Main, to, fmts, delete_from_library, subject='', send_ids=None, do_auto_convert=True, specific_format=None):
        ids = [self.library_view._model.id(r) for r in self.library_view.selectionModel().selectedRows()] if send_ids is None else send_ids
        if not ids or len(ids) == 0:
            return

        modified_metadata = []
        files, _auto_ids = self.library_view._model.get_preferred_formats_from_ids(
            ids,
            fmts,
            set_metadata=True,
            specific_format=specific_format,
            exclude_auto=do_auto_convert,
            use_plugboard=plugboard_email_value,
            plugboard_formats=plugboard_email_formats,
            modified_metadata=modified_metadata,
        )
        if do_auto_convert:
            nids = list(set(ids).difference(_auto_ids))
            ids = [i for i in ids if i in nids]
        else:
            _auto_ids = []

        full_metadata = self.library_view._model.metadata_for(ids, get_cover=False)

        bad, remove_ids, jobnames = [], [], []
        texts, subjects, attachments, attachment_names = [], [], [], []
        for f, mi, id, newmi in zip(files, full_metadata, ids, modified_metadata):
            if not newmi:
                newmi = mi
            t = mi.title or _('Unknown')
            if f is None:
                bad.append(t)
            else:
                remove_ids.append(id)
                jobnames.append(t)
                attachments.append(f)
                if not subject:
                    subjects.append(_('E-book:') + ' ' + t)
                else:
                    components = get_components(subject, mi, id)
                    if not components:
                        components = [mi.title]
                    subjects.append(os.path.join(*components))
                a = authors_to_string(mi.authors or [_('Unknown')])
                texts.append(
                    _('Attached, you will find the e-book')
                    + '\n\n'
                    + t
                    + '\n\t'
                    + _('by')
                    + ' '
                    + a
                    + '\n\n'
                    + _('in the %s format.') % os.path.splitext(f)[1][1:].upper()
                )
                if mi.comments and gprefs['add_comments_to_email']:
                    from calibre.ebooks.metadata import fmt_sidx
                    from calibre.utils.html2text import html2text

                    if mi.series:
                        sidx = fmt_sidx(
                            1.0 if mi.series_index is None else mi.series_index,
                            use_roman=config['use_roman_numerals_for_series_number'],
                        )
                        texts[-1] += '\n\n' + _('{series_index} of {series}').format(series_index=sidx, series=mi.series)
                    texts[-1] += '\n\n' + _('About this book:') + '\n\n' + textwrap.fill(html2text(mi.comments))
                if is_for_kindle(to):
                    prefix = str(newmi.title or t)
                else:
                    prefix = f'{t} - {a}'
                if not isinstance(prefix, str):
                    prefix = prefix.decode(preferred_encoding, 'replace')
                attachment_names.append(prefix + os.path.splitext(f)[1])
        remove = remove_ids if delete_from_library else []

        to_s = list(repeat(to, len(attachments)))
        if attachments:
            send_mails(
                jobnames,
                Dispatcher(partial(self.email_sent, remove=remove)),
                attachments,
                to_s,
                subjects,
                texts,
                attachment_names,
                self.job_manager,
            )
            self.status_bar.show_message(_('Sending email to') + ' ' + to, 3000)

        auto = []
        db = self.current_db.new_api
        if _auto_ids != []:
            for id in _auto_ids:
                if specific_format is None:
                    dbfmts = self.current_db.formats(id, index_is_id=True)
                    formats = [f.lower() for f in (dbfmts.split(',') if dbfmts else [])]
                    if set(formats).intersection(available_input_formats()) and set(fmts).intersection(available_output_formats()):
                        auto.append(id)
                    else:
                        bad.append(db.field_for('title', id))
                elif specific_format in list(set(fmts).intersection(set(available_output_formats()))):
                    auto.append(id)
                else:
                    bad.append(db.field_for('title', id))

        if auto != []:
            format = specific_format if specific_format in list(set(fmts).intersection(set(available_output_formats()))) else None
            if not format:
                for fmt in fmts:
                    if fmt in list(set(fmts).intersection(set(available_output_formats()))):
                        format = fmt
                        break
            if format is None:
                bad += auto
            else:
                autos = [db.field_for('title', id) for id in auto]
                if self.auto_convert_question(_('Auto convert the following books to %s before sending via email?') % format.upper(), autos):
                    self.iactions['Convert Books'].auto_convert_mail(to, fmts, delete_from_library, auto, format, subject)

        if bad:
            bad = '\n'.join(f'{i}' for i in bad)
            d = warning_dialog(
                self,
                _('No suitable formats'),
                _('Could not email the following books as no suitable formats were found:'),
                bad,
            )
            d.exec()

    def email_sent(self: Main, job, remove=[]):
        if job.failed:
            self.job_exception(job, dialog_title=_('Failed to email book'))
            return

        self.status_bar.show_message(job.description + ' ' + _('sent'), 5000)
        if remove:
            try:
                next_id = self.library_view.next_id
                self.library_view._model.delete_books_by_id(remove)
                self.iactions['Remove Books'].library_ids_deleted2(remove, next_id=next_id)
            except Exception:
                import traceback

                # Probably the user deleted the files, in any case, failing
                # to delete the book is not catastrophic
                traceback.print_exc()

    def email_news(self: Main, id_):
        mi = self.current_db.get_metadata(id_, index_is_id=True)
        remove = [id_] if config['delete_news_from_library_on_upload'] else []

        def get_fmts(fmts):
            files, auto = self.library_view._model.get_preferred_formats_from_ids(
                [id_],
                fmts,
                set_metadata=True,
                use_plugboard=plugboard_email_value,
                plugboard_formats=plugboard_email_formats,
            )
            return files

        sent_mails = email_news(mi, remove, get_fmts, self.email_sent, self.job_manager)
        if sent_mails:
            self.status_bar.show_message(_('Sent news to') + ' ' + ', '.join(sent_mails), 3000)


# }}}

if __name__ == '__main__':
    from qt.core import QApplication

    app = QApplication([])  # noqa: F841
    print(select_recipients())



#!/usr/bin/env python
# License: GPL v3 Copyright: 2019, Kovid Goyal <kovid at kovidgoyal.net>

import errno
import json
import os
import tarfile
from functools import lru_cache

from calibre.constants import cache_dir
from calibre.ptempfile import TemporaryDirectory
from calibre.utils.localization import lang_as_iso639_1
from calibre.utils.lock import ExclusiveFile
from calibre.utils.resources import get_path as P

_locale_map_cache: dict | None = None


def locale_map():
    global _locale_map_cache
    if _locale_map_cache is None:
        _locale_map_cache = {k.lower(): v for k, v in json.loads(P('hyphenation/locales.json', data=True)).items()}
    return _locale_map_cache


@lru_cache
def dictionary_name_for_locale(loc):
    loc = loc.lower().replace('-', '_')
    lmap = locale_map()
    if loc in lmap:
        return lmap[loc]
    parts = loc.split('_')
    if len(parts) > 2:
        loc = '_'.join(parts[:2])
        if loc in lmap:
            return lmap[loc]
    loc = lang_as_iso639_1(parts[0])
    if not loc:
        return
    if loc in lmap:
        return lmap[loc]
    if loc == 'en':
        return lmap['en_us']
    if loc == 'de':
        return lmap['de_de']
    if loc == 'es':
        return lmap['es_es']
    q = loc + '_'
    for k, v in lmap.items():
        if k.startswith(q):
            return v


@lru_cache(maxsize=2)
def expected_hash():
    return P('hyphenation/sha1sum', data=True, allow_user_override=False)


def extract_dicts(cache_path):
    dict_tarball = P('hyphenation/dictionaries.tar.xz', allow_user_override=False)
    with TemporaryDirectory(dir=cache_path) as tdir:
        tf = tarfile.open(dict_tarball)
        with tf:
            try:
                tf.extractall(tdir, filter='data')
            except TypeError:
                tf.extractall(tdir)
        with open(os.path.join(tdir, 'sha1sum'), 'wb') as f:
            f.write(expected_hash())
        dest = os.path.join(cache_path, 'f')
        with TemporaryDirectory(dir=cache_path) as trash:
            try:
                os.rename(dest, os.path.join(trash, 'f'))
            except OSError as err:
                if err.errno != errno.ENOENT:
                    raise
            os.rename(tdir, dest)
    is_cache_up_to_date.updated = True


class _IsCacheUpToDate:
    updated: bool = False

    def __call__(self, cache_path: str) -> bool:
        if self.updated:
            return True
        try:
            with open(os.path.join(cache_path, 'f', 'sha1sum'), 'rb') as f:
                actual_hash = f.read()
            if actual_hash == expected_hash():
                self.updated = True
                return True
        except OSError:
            pass
        return False


is_cache_up_to_date = _IsCacheUpToDate()


@lru_cache
def get_cache_path(cd):
    cache_path = os.path.join(cd, 'hyphenation')
    try:
        os.makedirs(cache_path)
    except OSError as err:
        if err.errno != errno.EEXIST:
            raise
    return cache_path


def path_to_dictionary(dictionary_name, cache_callback=None):
    cd = getattr(path_to_dictionary, 'cache_dir', None) or cache_dir()
    cache_path = get_cache_path(cd)
    with ExclusiveFile(os.path.join(cache_path, 'lock')):
        if not is_cache_up_to_date(cache_path):
            extract_dicts(cache_path)
            if cache_callback is not None:
                cache_callback()
    return os.path.join(cache_path, 'f', dictionary_name)



# License: GPLv3 Copyright: 2009, John Schember <john@nachtimwald.com>

import os
from optparse import OptionParser

from calibre.constants import iswindows
from calibre.customize.conversion import DummyReporter, OptionRecommendation
from calibre.customize.ui import plugin_for_catalog_format
from calibre.ebooks.conversion.plumber import Plumber
from calibre.utils.logging import Log


def gui_convert(
    input,
    output,
    recommendations,
    notification=DummyReporter(),
    abort_after_input_dump=False,
    log=None,
    override_input_metadata=False,
):
    recommendations = list(recommendations)
    recommendations.append(('verbose', 2, OptionRecommendation.HIGH))
    if log is None:
        log = Log()
    plumber = Plumber(
        input,
        output,
        log,
        report_progress=notification,
        abort_after_input_dump=abort_after_input_dump,
        override_input_metadata=override_input_metadata,
    )
    plumber.merge_ui_recommendations(recommendations)

    plumber.run()


def gui_convert_recipe(
    input,
    output,
    recommendations,
    notification=DummyReporter(),
    abort_after_input_dump=False,
    log=None,
    override_input_metadata=False,
):
    os.environ['CALIBRE_RECIPE_URN'] = input
    gui_convert(
        'from-gui.recipe',
        output,
        recommendations,
        notification=notification,
        abort_after_input_dump=abort_after_input_dump,
        log=log,
        override_input_metadata=override_input_metadata,
    )


def gui_convert_override(input, output, recommendations, notification=DummyReporter(), abort_after_input_dump=False, log=None):
    gui_convert(
        input,
        output,
        recommendations,
        notification=notification,
        abort_after_input_dump=abort_after_input_dump,
        log=log,
        override_input_metadata=True,
    )


def gui_catalog(
    library_path,
    temp_db_path,
    fmt,
    title,
    dbspec,
    ids,
    out_file_name,
    sync,
    fmt_options,
    connected_device,
    notification=DummyReporter(),
    log=None,
):
    if log is None:
        log = Log()
    from calibre.utils.config import prefs

    prefs.refresh()

    # Open the temp database created while still in the GUI thread
    from calibre.db.legacy import LibraryDatabase

    db = LibraryDatabase(library_path, temp_db_path=temp_db_path)
    try:
        setattr(db, 'catalog_plugin_on_device_temp_mapping', dbspec)

        # Create a minimal OptionParser that we can append to
        parser = OptionParser()
        args = []
        parser.add_option('--verbose', action='store_true', dest='verbose', default=True)
        opts, args = parser.parse_args()

        # Populate opts
        # opts.gui_search_text = something
        opts.catalog_title = title
        opts.connected_device = connected_device
        opts.ids = ids
        opts.search_text = None
        opts.sort_by = None
        opts.sync = sync

        # Extract the option dictionary to comma-separated lists
        for option in fmt_options:
            if isinstance(fmt_options[option], list):
                setattr(opts, option, ','.join(fmt_options[option]))
            else:
                setattr(opts, option, fmt_options[option])

        # Fetch and run the plugin for fmt
        # Returns 0 if successful, 1 if no catalog built
        plugin = plugin_for_catalog_format(fmt)
        return plugin.run(out_file_name, opts, db, notification=notification)
    finally:
        db.close()
        from calibre.db.backend import rmtree_with_retry

        try:
            rmtree_with_retry(os.path.dirname(temp_db_path))
        except PermissionError:
            if not iswindows:  # probably some antivirus holding a file open, the folder will be deleted on exit anyway
                raise



# License: GPLv3 Copyright: 2008, Marshall T. Vandegrift <llasram@gmail.com>

"""
CSS property propagation class.
"""

import copy
import logging
import numbers
import os
import re
import unicodedata
from operator import itemgetter
from typing import cast
from weakref import WeakKeyDictionary
from xml.dom import SyntaxErr as CSSSyntaxError

from css_parser import CSSParser, parseString, parseStyle, profiles, replaceUrls
from css_parser import log as css_parser_log
from css_parser import profile as cssprofiles
from css_parser.css import CSSFontFaceRule, CSSPageRule, CSSStyleRule, cssproperties
from css_selectors import INAPPROPRIATE_PSEUDO_CLASSES, Select, SelectorError
from tinycss.media3 import CSSMedia3Parser

from calibre import as_unicode, force_unicode
from calibre.ebooks import unit_convert
from calibre.ebooks.oeb.base import CSS_MIME, OEB_STYLES, SVG, XHTML, XHTML_NS, urlnormalize, xpath
from calibre.ebooks.oeb.normalize_css import DEFAULTS, normalizers
from calibre.utils.resources import get_path as P

css_parser_log.setLevel(logging.WARN)

_html_css_stylesheet = None


def validate_color(col):
    return cssprofiles.validateWithProfile('color', col, profiles=[profiles.Profiles.CSS_LEVEL_2])[1]


def html_css_stylesheet():
    global _html_css_stylesheet
    if _html_css_stylesheet is None:
        with open(P('templates/html.css'), 'rb') as f:
            html_css = f.read().decode('utf-8')
        _html_css_stylesheet = parseString(html_css, validate=False)
    return _html_css_stylesheet


INHERITED = {
    'azimuth',
    'border-collapse',
    'border-spacing',
    'caption-side',
    'color',
    'cursor',
    'direction',
    'elevation',
    'empty-cells',
    'font-family',
    'font-size',
    'font-style',
    'font-variant',
    'font-weight',
    'letter-spacing',
    'line-height',
    'list-style-image',
    'list-style-position',
    'list-style-type',
    'orphans',
    'page-break-inside',
    'pitch-range',
    'pitch',
    'quotes',
    'richness',
    'speak-header',
    'speak-numeral',
    'speak-punctuation',
    'speak',
    'speech-rate',
    'stress',
    'text-align',
    'text-indent',
    'text-transform',
    'visibility',
    'voice-family',
    'volume',
    'white-space',
    'widows',
    'word-spacing',
    'text-shadow',
}

FONT_SIZE_NAMES = {'xx-small', 'x-small', 'small', 'medium', 'large', 'x-large', 'xx-large'}

ALLOWED_MEDIA_TYPES = frozenset({'screen', 'all', 'aural', 'amzn-kf8'})
IGNORED_MEDIA_FEATURES = frozenset(
    'width min-width max-width height min-height max-height device-width min-device-width max-device-width device-height min-device-height max-device-height aspect-ratio min-aspect-ratio max-aspect-ratio device-aspect-ratio min-device-aspect-ratio max-device-aspect-ratio color min-color max-color color-index min-color-index max-color-index monochrome min-monochrome max-monochrome -webkit-min-device-pixel-ratio resolution min-resolution max-resolution scan grid'.split()  # noqa: E501
)


def media_ok(raw):
    if not raw:
        return True
    if raw == 'amzn-mobi':  # Optimization for the common case
        return False

    def query_ok(mq):
        matched = True
        if mq.media_type not in ALLOWED_MEDIA_TYPES:
            matched = False
        # Media queries that test for device specific features always fail
        for media_feature, expr in mq.expressions:
            if media_feature in IGNORED_MEDIA_FEATURES:
                matched = False
        return mq.negated ^ matched

    try:
        for mq in CSSMedia3Parser().parse_stylesheet(f'@media {raw} {{}}').rules[0].media:
            if query_ok(mq):
                return True
        return False
    except Exception:
        pass
    return True


def test_media_ok():
    assert media_ok(None)
    assert media_ok('')
    assert not media_ok('amzn-mobi')
    assert media_ok('amzn-kf8')
    assert media_ok('screen')
    assert media_ok('only screen')
    assert not media_ok('not screen')
    assert not media_ok('(device-width:10px)')
    assert media_ok('screen, (device-width:10px)')
    assert not media_ok('screen and (device-width:10px)')


class style_map(dict):
    def __init__(self):
        super().__init__()
        self.important_properties = set()


def epub_prefix_properties() -> dict[str, tuple[str, str]]:
    ans = getattr(epub_prefix_properties, 'ans', None)
    if ans is None:
        ans = {}
        for unprefixed in (
            'writing-mode',
            'text-emphasis',
            'text-emphasis-color',
            'text-emphasis-position',
            'text-emphasis-style',
        ):
            ans[f'-epub-{unprefixed}'] = f'-webkit-{unprefixed}', unprefixed
        setattr(epub_prefix_properties, 'ans', ans)
    return ans


def cleanup_epub_prefixed_properties(style: dict[str, str]) -> None:
    epub_prefixed = epub_prefix_properties()
    for x in tuple(filter(epub_prefixed.__contains__, style)):
        val = style.pop(x)
        for key in epub_prefixed[x]:
            if key not in style:
                style[key] = val


class StylizerRules:
    def __init__(self, opts, profile, stylesheets):
        self.opts, self.profile, self.stylesheets = opts, profile, stylesheets

        index = 0
        self.rules = []
        self.page_rule = {}
        self.font_face_rules = []
        for sheet_index, stylesheet in enumerate(stylesheets):
            href = stylesheet.href
            for rule in stylesheet.cssRules:
                if rule.type == rule.MEDIA_RULE:
                    if media_ok(rule.media.mediaText):
                        for subrule in rule.cssRules:
                            self.rules.extend(self.flatten_rule(subrule, href, index, is_user_agent_sheet=sheet_index == 0))
                            index += 1
                else:
                    self.rules.extend(self.flatten_rule(rule, href, index, is_user_agent_sheet=sheet_index == 0))
                    index = index + 1
        self.rules.sort(key=itemgetter(0))  # sort by specificity

    def flatten_rule(self, rule, href, index, is_user_agent_sheet=False):
        results = []
        sheet_index = 0 if is_user_agent_sheet else 1
        if isinstance(rule, CSSStyleRule):
            style = self.flatten_style(rule.style)
            for selector in rule.selectorList:
                specificity = (sheet_index,) + selector.specificity + (index,)
                text = selector.selectorText
                selector = list(selector.seq)
                results.append((specificity, selector, style, text, href))
        elif isinstance(rule, CSSPageRule):
            style = self.flatten_style(rule.style)
            self.page_rule.update(style)
        elif isinstance(rule, CSSFontFaceRule):
            if rule.style.length > 1:
                # Ignore the meaningless font face rules generated by the
                # benighted MS Word that contain only a font-family declaration
                # and nothing else
                self.font_face_rules.append(rule)
        return results

    def flatten_style(self, cssstyle):
        style = style_map()
        for prop in cssstyle:
            name = prop.name
            normalizer = normalizers.get(name, None)
            is_important = prop.priority == 'important'
            if normalizer is not None:
                for name, val in normalizer(name, prop.propertyValue).items():
                    style[name] = val
                    if is_important:
                        style.important_properties.add(name)
            elif name == 'text-align':
                style['text-align'] = self._apply_text_align(prop.value)
                if is_important:
                    style.important_properties.add(name)
            else:
                style[name] = prop.value
                if is_important:
                    style.important_properties.add(name)
        if 'font-size' in style:
            size = style['font-size']
            if size == 'normal':
                size = 'medium'
            if size == 'smallest':
                size = 'xx-small'
            if size in FONT_SIZE_NAMES:
                style['font-size'] = f'{self.profile.fnames[size] / float(self.profile.fbase):.1f}rem'
        cleanup_epub_prefixed_properties(style)
        return style

    def _apply_text_align(self, text):
        if text in ('left', 'justify') and self.opts.change_justification in ('left', 'justify'):
            text = self.opts.change_justification
        return text

    def same_rules(self, opts, profile, stylesheets):
        if self.opts != opts:
            # it's unlikely to happen, but better safe than sorry
            return False
        if self.profile != profile:
            return False
        if len(self.stylesheets) != len(stylesheets):
            return False
        for index, stylesheet in enumerate(self.stylesheets):
            if stylesheet != stylesheets[index]:
                return False
        return True


class Stylizer:
    STYLESHEETS = WeakKeyDictionary()

    def __init__(self, tree, path, oeb, opts, profile=None, extra_css='', user_css='', base_css=''):
        self.oeb, self.opts = oeb, opts
        self.profile = profile
        if self.profile is None:
            # Use the default profile. This should really be using
            # opts.output_profile, but I don't want to risk changing it, as
            # doing so might well have hard to debug font size effects.
            from calibre.customize.ui import output_profiles

            for x in output_profiles():
                if x.short_name == 'default':
                    self.profile = x
                    break
        if self.profile is None:
            # Just in case the default profile is removed in the future :)
            self.profile = opts.output_profile
        self.body_font_size = self.profile.fbase
        self.logger = oeb.logger
        item = oeb.manifest.hrefs[path]
        basename = os.path.basename(path)
        cssname = os.path.splitext(basename)[0] + '.css'
        stylesheets = [html_css_stylesheet()]
        if base_css:
            stylesheets.append(parseString(base_css, validate=False))
        style_tags = xpath(tree, '//*[local-name()="style" or local-name()="link"]')

        # Add css_parser parsing profiles from output_profile
        for profile in self.opts.output_profile.extra_css_modules:
            cssprofiles.addProfile(profile['name'], profile['props'], profile['macros'])

        parser = CSSParser(fetcher=self._fetch_css_file, log=logging.getLogger('calibre.css'))
        for elem in style_tags:
            if elem.tag in (XHTML('style'), SVG('style')) and elem.get('type', CSS_MIME) in OEB_STYLES and media_ok(elem.get('media')):
                text = elem.text or ''
                for x in elem:
                    t = getattr(x, 'text', None)
                    if t:
                        text += '\n\n' + force_unicode(t, 'utf-8')
                    t = getattr(x, 'tail', None)
                    if t:
                        text += '\n\n' + force_unicode(t, 'utf-8')
                if text:
                    text = oeb.css_preprocessor(text)
                    # We handle @import rules separately
                    parser.setFetcher(lambda x: ('utf-8', b''))
                    stylesheet = parser.parseString(text, href=cssname, validate=False)
                    parser.setFetcher(self._fetch_css_file)
                    for rule in stylesheet.cssRules:
                        if rule.type == rule.IMPORT_RULE:
                            ihref = item.abshref(rule.href)
                            if not media_ok(rule.media.mediaText):
                                continue
                            hrefs = self.oeb.manifest.hrefs
                            if ihref not in hrefs:
                                self.logger.warn('Ignoring missing stylesheet in @import rule:', rule.href)
                                continue
                            sitem = hrefs[ihref]
                            if sitem.media_type not in OEB_STYLES:
                                self.logger.warn(f'CSS @import of non-CSS file {rule.href!r}')
                                continue
                            stylesheets.append(sitem.data)
                    # Make links to resources absolute, since these rules will
                    # be folded into a stylesheet at the root
                    replaceUrls(stylesheet, item.abshref, ignoreImportRules=True)
                    stylesheets.append(stylesheet)
            elif (
                elem.tag == XHTML('link')
                and elem.get('href')
                and elem.get('rel', 'stylesheet').lower() == 'stylesheet'
                and elem.get('type', CSS_MIME).lower() in OEB_STYLES
                and media_ok(elem.get('media'))
            ):
                href = urlnormalize(elem.attrib['href'])
                path = item.abshref(href)
                sitem = oeb.manifest.hrefs.get(path, None)
                if sitem is None:
                    self.logger.warn(f'Stylesheet {path!r} referenced by file {item.href!r} not in manifest')
                    continue
                if not hasattr(sitem.data, 'cssRules'):
                    self.logger.warn(f'Stylesheet {path!r} referenced by file {item.href!r} is not CSS')
                    continue
                stylesheets.append(sitem.data)
        csses = {'extra_css': extra_css, 'user_css': user_css}
        for w, x in csses.items():
            if x:
                try:
                    text = x
                    stylesheet = parser.parseString(text, href=cssname, validate=False)
                    stylesheets.append(stylesheet)
                except Exception:
                    self.logger.exception(f'Failed to parse {w}, ignoring.')
                    self.logger.debug('Bad css: ')
                    self.logger.debug(x)

        # using oeb to store the rules, page rule and font face rules
        # and generating them again if opts, profile or stylesheets are different
        if (not hasattr(self.oeb, 'stylizer_rules')) or not self.oeb.stylizer_rules.same_rules(self.opts, self.profile, stylesheets):
            self.oeb.stylizer_rules = StylizerRules(self.opts, self.profile, stylesheets)
        self.rules = self.oeb.stylizer_rules.rules
        self.page_rule = self.oeb.stylizer_rules.page_rule
        self.font_face_rules = self.oeb.stylizer_rules.font_face_rules
        self.flatten_style = self.oeb.stylizer_rules.flatten_style

        self._styles = {}
        pseudo_pat = re.compile(':{{1,2}}({})'.format('|'.join(INAPPROPRIATE_PSEUDO_CLASSES)), re.I)
        select = Select(tree, ignore_inappropriate_pseudo_classes=True)

        for _, _, cssdict, text, _ in self.rules:
            fl = pseudo_pat.search(text)
            try:
                matches = tuple(select(text))
            except SelectorError as err:
                self.logger.error(f'Ignoring CSS rule with invalid selector: {text!r} ({as_unicode(err)})')
                continue

            if fl is not None:
                fl = fl.group(1)
                if fl == 'first-letter' and getattr(self.oeb, 'plumber_output_format', '').lower() in {'mobi', 'docx'}:
                    # Fake first-letter
                    for elem in matches:
                        for x in elem.iter('*'):
                            if x.text:
                                punctuation_chars = []
                                text = str(x.text)
                                while text:
                                    category = unicodedata.category(text[0])
                                    if category[0] not in {'P', 'Z'}:
                                        break
                                    punctuation_chars.append(text[0])
                                    text = text[1:]

                                special_text = ''.join(punctuation_chars) + (text[0] if text else '')
                                span = x.makeelement(f'{{{XHTML_NS}}}span')
                                span.text = special_text
                                span.set('data-fake-first-letter', '1')
                                span.tail = text[1:]
                                x.text = None
                                x.insert(0, span)
                                self.style(span)._update_cssdict(cssdict)
                                break
                else:  # Element pseudo-class
                    for elem in matches:
                        self.style(elem)._update_pseudo_class(fl, cssdict)
            else:
                for elem in matches:
                    self.style(elem)._update_cssdict(cssdict)
        for elem in xpath(tree, '//h:*[@style]'):
            self.style(elem)._apply_style_attr(url_replacer=item.abshref)
        num_pat = re.compile(r'[0-9.]+$')
        for elem in xpath(tree, '//h:img[@width or @height]'):
            style = self.style(elem)
            # Check if either height or width is not default
            is_styled = style._style.get('width', 'auto') != 'auto' or style._style.get('height', 'auto') != 'auto'
            if not is_styled:
                # Update img style dimension using width and height
                upd = {}
                for prop in ('width', 'height'):
                    val = elem.get(prop, '').strip()
                    try:
                        del elem.attrib[prop]
                    except Exception:
                        pass
                    if val:
                        if num_pat.match(val) is not None:
                            val += 'px'
                        upd[prop] = val
                if upd:
                    style._update_cssdict(upd)

    def _fetch_css_file(self, path):
        hrefs = self.oeb.manifest.hrefs
        if path not in hrefs:
            self.logger.warn(f'CSS import of missing file {path!r}')
            return None, None
        item = hrefs[path]
        if item.media_type not in OEB_STYLES:
            self.logger.warn(f'CSS import of non-CSS file {path!r}')
            return None, None
        data = item.data.cssText
        if not isinstance(data, bytes):
            data = data.encode('utf-8')
        return 'utf-8', data

    def style(self, element):
        try:
            return self._styles[element]
        except KeyError:
            return Style(element, self)

    def stylesheet(self, name, font_scale=None):
        rules = []
        for _, _, style, selector, href in self.rules:
            if href != name:
                continue
            if font_scale and 'font-size' in style and style['font-size'].endswith('pt'):
                style = copy.copy(style)
                size = float(style['font-size'][:-2])
                style['font-size'] = f'{size * font_scale:.2f}pt'
            style = ';\n    '.join(': '.join(item) for item in style.items())
            rules.append(f'{selector} {{\n    {style};\n}}')
        return '\n'.join(rules)


no_important_properties = frozenset()
svg_text_tags = tuple(map(SVG, ('text', 'textPath', 'tref', 'tspan')))


def is_only_number(x: str) -> bool:
    try:
        float(x)
        return True
    except Exception:
        return False


def is_svg_text_tag(x):
    return getattr(x, 'tag', '') in svg_text_tags


class Style:
    MS_PAT = re.compile(r'^\s*(mso-|panose-|text-underline|tab-interval)')
    viewport_relative_font_size: str = ''

    def __init__(self, element, stylizer):
        self._element = element
        self._profile = stylizer.profile
        self._stylizer = stylizer
        self._style = style_map()
        self._fontSize = None
        self._width = None
        self._height = None
        self._lineHeight = None
        self._bgcolor = None
        self._fgcolor = None
        self._pseudo_classes = {}
        stylizer._styles[element] = self

    def set(self, prop, val):
        self._style[prop] = val

    def drop(self, prop, default=None):
        return self._style.pop(prop, default)

    def _update_cssdict(self, cssdict):
        self._update_style(cssdict)

    def _update_style(self, cssdict):
        current_ip = getattr(self._style, 'important_properties', no_important_properties)
        if current_ip is no_important_properties:
            s = style_map()
            s.update(self._style)
            self._style = s
            current_ip = self._style.important_properties
        update_ip = getattr(cssdict, 'important_properties', no_important_properties)
        for name, val in cssdict.items():
            override = False
            if name in update_ip:
                cast(set, current_ip).add(name)
                override = True
            elif name not in current_ip:
                override = True
            if override:
                self._style[name] = val

    def _update_pseudo_class(self, name, cssdict):
        orig = self._pseudo_classes.get(name, {})
        orig.update(cssdict)
        self._pseudo_classes[name] = orig

    def _apply_style_attr(self, url_replacer=None):
        attrib = self._element.attrib
        if 'style' not in attrib:
            return
        css = attrib['style'].split(';')
        css = filter(None, (x.strip() for x in css))
        css = [y.strip() for y in css]
        css = [y for y in css if self.MS_PAT.match(y) is None]
        css = '; '.join(css)
        try:
            style = parseStyle(css, validate=False)
        except CSSSyntaxError:
            return
        if url_replacer is not None:
            replaceUrls(style, url_replacer, ignoreImportRules=True)
        self._update_style(self._stylizer.flatten_style(style))

    def _has_parent(self):
        try:
            return self._element.getparent() is not None
        except AttributeError:
            return False  # self._element is None

    def _get_parent(self):
        elem = self._element.getparent()
        if elem is None:
            return None
        return self._stylizer.style(elem)

    def __getitem__(self, name):
        domname = cssproperties._toDOMname(name)
        if hasattr(self, domname):
            return getattr(self, domname)
        return self._unit_convert(self._get(name))

    def _get(self, name):
        result = self._style.get(name, None)
        if result == 'inherit' or (result is None and name in INHERITED and self._has_parent()):
            stylizer = self._stylizer
            result = stylizer.style(self._element.getparent())._get(name)
        if result is None:
            result = DEFAULTS[name]
        return result

    def get(self, name, default=None):
        return self._style.get(name, default)

    def _unit_convert(self, value, base=None, font=None):
        "Return value in pts"
        if base is None:
            base = self.width
        if not font and font != 0:
            font = self.fontSize
        return unit_convert(value, base, font, self._profile.dpi, body_font_size=self._stylizer.body_font_size)

    def pt_to_px(self, value):
        return (self._profile.dpi / 72) * value

    @property
    def color(self):
        if self._fgcolor is None:
            val = self._get('color')
            if val and validate_color(val):
                self._fgcolor = val
            else:
                self._fgcolor = DEFAULTS['color']
        return self._fgcolor

    @property
    def backgroundColor(self):
        """
        Return the background color by parsing both the background-color and
        background shortcut properties. Note that inheritance/default values
        are not used. None is returned if no background color is set.
        """

        if self._bgcolor is None:
            col = None
            val = self._style.get('background-color', None)
            if val and validate_color(val):
                col = val
            else:
                val = self._style.get('background', None)
                if val is not None:
                    try:
                        style = parseStyle('background: ' + val, validate=False)
                        css_val = style.getProperty('background').propertyValue
                        try:
                            val_list = list(css_val)
                        except Exception:
                            # css_val is CSSPrimitiveValue
                            val_list = [css_val]
                        for c in val_list:
                            c = c.cssText
                            if isinstance(c, bytes):
                                c = c.decode('utf-8', 'replace')
                            if validate_color(c):
                                col = c
                                break
                    except Exception:
                        pass
            if col is None:
                self._bgcolor = False
            else:
                self._bgcolor = col
        return self._bgcolor or None

    @property
    def fontSize(self):
        def normalize_fontsize(value, base):
            value = value.replace('"', '').replace("'", '')
            result = None
            factor = None
            if value == 'inherit':
                value = base
            if value in FONT_SIZE_NAMES:
                result = self._profile.fnames[value]
            elif value == 'smaller':
                factor = 1.0 / 1.2
                for _, _, size in self._profile.fsizes:
                    if base <= size:
                        break
                    factor = None
                    result = size
            elif value == 'larger':
                factor = 1.2
                for _, _, size in reversed(self._profile.fsizes):
                    if base >= size:
                        break
                    factor = None
                    result = size
            else:
                result = self._unit_convert(value, base=base, font=base)
                if not isinstance(result, numbers.Number):
                    return base
                if result < 0:
                    result = normalize_fontsize('smaller', base)
            if factor:
                result = factor * base
            return result

        if self._fontSize is None:
            result = None
            parent = self._get_parent()
            if parent is not None:
                base = parent.fontSize
            else:
                base = self._profile.fbase
            if 'font-size' in self._style:
                size = self._style['font-size']
                if is_svg_text_tag(self._element) and (size.endswith('px') or is_only_number(size)):
                    self.viewport_relative_font_size = size
                result = normalize_fontsize(size, base)
            else:
                result = base
            self._fontSize = result
        return self._fontSize

    def img_dimension(self, attr, img_size):
        ans = None
        parent = self._get_parent()
        if parent is not None:
            base = getattr(parent, attr)
        else:
            base = getattr(self._profile, attr + '_pts')
        x = self._style.get(attr)
        if x is not None:
            if x == 'auto':
                ans = self._unit_convert(str(img_size) + 'px', base=base)
            else:
                x = self._unit_convert(x, base=base)
                if isinstance(x, numbers.Number):
                    ans = x
        if ans is None:
            x = self._element.get(attr)
            if x is not None:
                x = self._unit_convert(x + 'px', base=base)
                if isinstance(x, numbers.Number):
                    ans = x
        if ans is None:
            ans = self._unit_convert(str(img_size) + 'px', base=base)
        maa = self._style.get('max-' + attr)
        if maa is not None:
            x = self._unit_convert(maa, base=base)
            if isinstance(x, numbers.Number) and (ans is None or x < ans):
                ans = x
        return ans

    def img_size(self, width, height):
        "Return the final size of an <img> given that it points to an image of size widthxheight"
        w, h = self._get('width'), self._get('height')
        answ, ansh = self.img_dimension('width', width), self.img_dimension('height', height)
        if w == 'auto' and h != 'auto':
            answ = (float(width) / height) * ansh
        elif h == 'auto' and w != 'auto':
            ansh = (float(height) / width) * answ
        return answ, ansh

    @property
    def width(self):
        if self._width is None:
            width = None
            base = None
            parent = self._get_parent()
            if parent is not None:
                base = parent.width
            else:
                base = self._profile.width_pts
            if 'width' in self._element.attrib:
                width = self._element.attrib['width']
            elif 'width' in self._style:
                width = self._style['width']
            if not width or width == 'auto':
                result = base
            else:
                result = self._unit_convert(width, base=base)
            if isinstance(result, (str, bytes)):
                result = self._profile.width
            self._width = result
            if 'max-width' in self._style:
                result = self._unit_convert(self._style['max-width'], base=base)
                if isinstance(result, (str, bytes)):
                    result = self._width
                self._width = min(self._width, result)

        return self._width

    @property
    def parent_width(self):
        parent = self._get_parent()
        if parent is None:
            return self.width
        return parent.width

    @property
    def height(self):
        if self._height is None:
            height = None
            base = None
            parent = self._get_parent()
            if parent is not None:
                base = parent.height
            else:
                base = self._profile.height_pts
            if 'height' in self._element.attrib:
                height = self._element.attrib['height']
            elif 'height' in self._style:
                height = self._style['height']
            if not height or height == 'auto':
                result = base
            else:
                result = self._unit_convert(height, base=base)
            if isinstance(result, (str, bytes)):
                result = self._profile.height
            self._height = result
            if 'max-height' in self._style:
                result = self._unit_convert(self._style['max-height'], base=base)
                if isinstance(result, (str, bytes)):
                    result = self._height
                self._height = min(self._height, result)

        return self._height

    @property
    def lineHeight(self):
        if self._lineHeight is None:
            result = None
            parent = self._get_parent()
            if 'line-height' in self._style:
                lineh = self._style['line-height']
                if lineh == 'normal':
                    lineh = '1.2'
                try:
                    result = float(lineh) * self.fontSize
                except ValueError:
                    result = self._unit_convert(lineh, base=self.fontSize)
            elif parent is not None:
                # TODO: proper inheritance
                result = parent.lineHeight
            else:
                result = 1.2 * self.fontSize
            self._lineHeight = result
        return self._lineHeight

    @property
    def effective_text_decoration(self):
        """
        Browsers do this creepy thing with text-decoration where even though the
        property is not inherited, it looks like it is because containing
        blocks apply it. The actual algorithm is utterly ridiculous, see
        http://reference.sitepoint.com/css/text-decoration
        This matters for MOBI output, where text-decoration is mapped to <u>
        and <st> tags. Trying to implement the actual algorithm is too much
        work, so we just use a simple fake that should cover most cases.
        """
        css = self._style.get('text-decoration', None)
        pcss = None
        parent = self._get_parent()
        if parent is not None:
            pcss = parent._style.get('text-decoration', None)
        if css in ('none', None, 'inherit') and pcss not in (None, 'none'):
            return pcss
        return css

    @property
    def first_vertical_align(self):
        """For docx output where tags are not nested, we cannot directly
        simulate the HTML vertical-align rendering model. Instead use the
        approximation of considering the first non-default vertical-align"""
        val = self['vertical-align']
        if val != 'baseline':
            raw_val = self._get('vertical-align')
            if '%' in raw_val:
                val = self._unit_convert(raw_val, base=self['line-height'])
            return val
        parent = self._get_parent()
        if parent is not None and 'inline' in parent['display']:
            return parent.first_vertical_align

    @property
    def marginTop(self):
        return self._unit_convert(self._get('margin-top'), base=self.parent_width)

    @property
    def marginBottom(self):
        return self._unit_convert(self._get('margin-bottom'), base=self.parent_width)

    @property
    def marginLeft(self):
        return self._unit_convert(self._get('margin-left'), base=self.parent_width)

    @property
    def marginRight(self):
        return self._unit_convert(self._get('margin-right'), base=self.parent_width)

    @property
    def paddingTop(self):
        return self._unit_convert(self._get('padding-top'), base=self.parent_width)

    @property
    def paddingBottom(self):
        return self._unit_convert(self._get('padding-bottom'), base=self.parent_width)

    @property
    def paddingLeft(self):
        return self._unit_convert(self._get('padding-left'), base=self.parent_width)

    @property
    def paddingRight(self):
        return self._unit_convert(self._get('padding-right'), base=self.parent_width)

    def __str__(self):
        items = sorted(self._style.items())
        return '; '.join(f'{key}: {val}' for key, val in items)

    def cssdict(self):
        return dict(self._style)

    def pseudo_classes(self, filter_css):
        if filter_css:
            css = copy.deepcopy(self._pseudo_classes)
            for psel, cssdict in css.items():
                for k in filter_css:
                    cssdict.pop(k, None)
        else:
            css = self._pseudo_classes
        return {k: v for k, v in css.items() if v}

    @property
    def is_hidden(self):
        return self._style.get('display') == 'none' or self._style.get('visibility') == 'hidden'



#!/usr/bin/env python
# License: GPLv3 Copyright: 2008, Kovid Goyal kovid@kovidgoyal.net

from struct import pack

from calibre.utils.localization import lang_as_iso639_1

lang_codes = {}

main_language = {
    0: 'NEUTRAL',
    54: 'AFRIKAANS',
    28: 'ALBANIAN',
    1: 'ARABIC',
    43: 'ARMENIAN',
    77: 'ASSAMESE',
    44: 'AZERI',
    45: 'BASQUE',
    35: 'BELARUSIAN',
    69: 'BENGALI',
    2: 'BULGARIAN',
    3: 'CATALAN',
    4: 'CHINESE',
    # 26 : "CROATIAN",
    5: 'CZECH',
    6: 'DANISH',
    19: 'DUTCH',
    9: 'ENGLISH',
    37: 'ESTONIAN',
    56: 'FAEROESE',
    41: 'FARSI',
    11: 'FINNISH',
    12: 'FRENCH',
    55: 'GEORGIAN',
    7: 'GERMAN',
    8: 'GREEK',
    71: 'GUJARATI',
    13: 'HEBREW',
    57: 'HINDI',
    14: 'HUNGARIAN',
    15: 'ICELANDIC',
    33: 'INDONESIAN',
    16: 'ITALIAN',
    17: 'JAPANESE',
    75: 'KANNADA',
    63: 'KAZAK',
    87: 'KONKANI',
    18: 'KOREAN',
    38: 'LATVIAN',
    39: 'LITHUANIAN',
    47: 'MACEDONIAN',
    62: 'MALAY',
    76: 'MALAYALAM',
    58: 'MALTESE',
    78: 'MARATHI',
    97: 'NEPALI',
    20: 'NORWEGIAN',
    72: 'ORIYA',
    21: 'POLISH',
    22: 'PORTUGUESE',
    70: 'PUNJABI',
    23: 'RHAETOROMANIC',
    24: 'ROMANIAN',
    25: 'RUSSIAN',
    59: 'SAMI',
    79: 'SANSKRIT',
    26: 'SERBIAN',
    27: 'SLOVAK',
    36: 'SLOVENIAN',
    46: 'SORBIAN',
    10: 'SPANISH',
    48: 'SUTU',
    65: 'SWAHILI',
    29: 'SWEDISH',
    73: 'TAMIL',
    68: 'TATAR',
    74: 'TELUGU',
    30: 'THAI',
    49: 'TSONGA',
    50: 'TSWANA',
    31: 'TURKISH',
    34: 'UKRAINIAN',
    32: 'URDU',
    67: 'UZBEK',
    42: 'VIETNAMESE',
    52: 'XHOSA',
    53: 'ZULU',
}

sub_language = {
    0: 'NEUTRAL',
    # 1 : "ARABIC_SAUDI_ARABIA",
    # 2 : "ARABIC_IRAQ",
    # 3 : "ARABIC_EGYPT",
    # 4 : "ARABIC_LIBYA",
    # 5 : "ARABIC_ALGERIA",
    # 6 : "ARABIC_MOROCCO",
    # 7 : "ARABIC_TUNISIA",
    # 8 : "ARABIC_OMAN",
    # 9 : "ARABIC_YEMEN",
    # 10 : "ARABIC_SYRIA",
    # 11 : "ARABIC_JORDAN",
    # 12 : "ARABIC_LEBANON",
    # 13 : "ARABIC_KUWAIT",
    # 14 : "ARABIC_UAE",
    # 15 : "ARABIC_BAHRAIN",
    # 16 : "ARABIC_QATAR",
    # 1 : "AZERI_LATIN",
    # 2 : "AZERI_CYRILLIC",
    # 1 : "CHINESE_TRADITIONAL",
    # 2 : "CHINESE_SIMPLIFIED",
    # 3 : "CHINESE_HONGKONG",
    # 4 : "CHINESE_SINGAPORE",
    # 1 : "DUTCH",
    # 2 : "DUTCH_BELGIAN",
    # 1 : "FRENCH",
    # 2 : "FRENCH_BELGIAN",
    # 3 : "FRENCH_CANADIAN",
    # 4 : "FRENCH_SWISS",
    # 5 : "FRENCH_LUXEMBOURG",
    # 6 : "FRENCH_MONACO",
    # 1 : "GERMAN",
    # 2 : "GERMAN_SWISS",
    # 3 : "GERMAN_AUSTRIAN",
    # 4 : "GERMAN_LUXEMBOURG",
    # 5 : "GERMAN_LIECHTENSTEIN",
    # 1 : "ITALIAN",
    # 2 : "ITALIAN_SWISS",
    # 1 : "KOREAN",
    # 1 : "LITHUANIAN",
    # 1 : "MALAY_MALAYSIA",
    # 2 : "MALAY_BRUNEI_DARUSSALAM",
    # 1 : "NORWEGIAN_BOKMAL",
    # 2 : "NORWEGIAN_NYNORSK",
    # 2 : "PORTUGUESE",
    # 1 : "PORTUGUESE_BRAZILIAN",
    # 2 : "SERBIAN_LATIN",
    3: 'SERBIAN_CYRILLIC',
    # 1 : "SPANISH",
    # 2 : "SPANISH_MEXICAN",
    4: 'SPANISH_GUATEMALA',
    5: 'SPANISH_COSTA_RICA',
    6: 'SPANISH_PANAMA',
    7: 'SPANISH_DOMINICAN_REPUBLIC',
    8: 'SPANISH_VENEZUELA',
    9: 'SPANISH_COLOMBIA',
    10: 'SPANISH_PERU',
    11: 'SPANISH_ARGENTINA',
    12: 'SPANISH_ECUADOR',
    13: 'SPANISH_CHILE',
    14: 'SPANISH_URUGUAY',
    15: 'SPANISH_PARAGUAY',
    16: 'SPANISH_BOLIVIA',
    17: 'SPANISH_EL_SALVADOR',
    18: 'SPANISH_HONDURAS',
    19: 'SPANISH_NICARAGUA',
    20: 'SPANISH_PUERTO_RICO',
    # 1 : "SWEDISH",
    # 2 : "SWEDISH_FINLAND",
    1: 'UZBEK_LATIN',
    2: 'UZBEK_CYRILLIC',
}

IANA_MOBI = {
    None: {None: (0, 0)},
    'af': {None: (54, 0)},
    'ar': {
        None: (1, 0),
        'AE': (1, 56),
        'BH': (1, 60),
        'DZ': (1, 20),
        'EG': (1, 12),
        'JO': (1, 44),
        'KW': (1, 52),
        'LB': (1, 48),
        'MA': (1, 24),
        'OM': (1, 32),
        'QA': (1, 64),
        'SA': (1, 4),
        'SY': (1, 40),
        'TN': (1, 28),
        'YE': (1, 36),
    },
    'as': {None: (77, 0)},
    'az': {None: (44, 0)},
    'be': {None: (35, 0)},
    'bg': {None: (2, 0)},
    'bn': {None: (69, 0)},
    'ca': {None: (3, 0)},
    'cs': {None: (5, 0)},
    'da': {None: (6, 0)},
    'de': {None: (7, 0), 'AT': (7, 12), 'CH': (7, 8), 'LI': (7, 20), 'LU': (7, 16)},
    'el': {None: (8, 0)},
    'en': {
        None: (9, 0),
        'AU': (9, 12),
        'BZ': (9, 40),
        'CA': (9, 16),
        'GB': (9, 8),
        'IE': (9, 24),
        'JM': (9, 32),
        'NZ': (9, 20),
        'PH': (9, 52),
        'TT': (9, 44),
        'US': (9, 4),
        'ZA': (9, 28),
        'ZW': (9, 48),
    },
    'es': {
        None: (10, 0),
        'AR': (10, 44),
        'BO': (10, 64),
        'CL': (10, 52),
        'CO': (10, 36),
        'CR': (10, 20),
        'DO': (10, 28),
        'EC': (10, 48),
        'ES': (10, 4),
        'GT': (10, 16),
        'HN': (10, 72),
        'MX': (10, 8),
        'NI': (10, 76),
        'PA': (10, 24),
        'PE': (10, 40),
        'PR': (10, 80),
        'PY': (10, 60),
        'SV': (10, 68),
        'UY': (10, 56),
        'VE': (10, 32),
    },
    'et': {None: (37, 0)},
    'eu': {None: (45, 0)},
    'fa': {None: (41, 0)},
    'fi': {None: (11, 0)},
    'fo': {None: (56, 0)},
    'fr': {None: (12, 0), 'BE': (12, 8), 'CA': (12, 12), 'CH': (12, 16), 'FR': (12, 4), 'LU': (12, 20), 'MC': (12, 24)},
    'gu': {None: (71, 0)},
    'he': {None: (13, 0)},
    'hi': {None: (57, 0)},
    'hr': {None: (26, 0)},
    'hu': {None: (14, 0)},
    'hy': {None: (43, 0)},
    'id': {None: (33, 0)},
    'is': {None: (15, 0)},
    'it': {None: (16, 0), 'CH': (16, 8), 'IT': (16, 4)},
    'ja': {None: (17, 0)},
    'ka': {None: (55, 0)},
    'kk': {None: (63, 0)},
    'kn': {None: (75, 0)},
    'ko': {None: (18, 0)},
    'kok': {None: (87, 0)},
    'lt': {None: (39, 0)},
    'lv': {None: (38, 0)},
    'mk': {None: (47, 0)},
    'ml': {None: (76, 0)},
    'mr': {None: (78, 0)},
    'ms': {None: (62, 0)},
    'mt': {None: (58, 0)},
    'ne': {None: (97, 0)},
    'nl': {None: (19, 0), 'BE': (19, 8)},
    'no': {None: (20, 0)},
    'or': {None: (72, 0)},
    'pa': {None: (70, 0)},
    'pl': {None: (21, 0)},
    'pt': {None: (22, 0), 'BR': (22, 4), 'PT': (22, 8)},
    'rm': {None: (23, 0)},
    'ro': {None: (24, 0)},
    'ru': {None: (25, 0)},
    'sa': {None: (79, 0)},
    'se': {None: (59, 0)},
    'sk': {None: (27, 0)},
    'sl': {None: (36, 0)},
    'sq': {None: (28, 0)},
    'sr': {None: (26, 12), 'RS': (26, 12)},
    'st': {None: (48, 0)},
    'sv': {None: (29, 0), 'FI': (29, 8)},
    'sw': {None: (65, 0)},
    'ta': {None: (73, 0)},
    'te': {None: (74, 0)},
    'th': {None: (30, 0)},
    'tn': {None: (50, 0)},
    'tr': {None: (31, 0)},
    'ts': {None: (49, 0)},
    'tt': {None: (68, 0)},
    'uk': {None: (34, 0)},
    'ur': {None: (32, 0)},
    'uz': {None: (67, 0), 'UZ': (67, 8)},
    'vi': {None: (42, 0)},
    'wen': {None: (46, 0)},
    'xh': {None: (52, 0)},
    'zh': {None: (4, 0), 'CN': (4, 8), 'HK': (4, 12), 'SG': (4, 16), 'TW': (4, 4)},
    'zu': {None: (53, 0)},
}


def iana2mobi(icode):
    langdict, subtags = IANA_MOBI[None], []
    if icode:
        subtags = list(icode.split('-'))
        while len(subtags) > 0:
            lang = subtags.pop(0).lower()
            lang = lang_as_iso639_1(lang)
            if lang and lang in IANA_MOBI:
                langdict = IANA_MOBI[lang]
                break

    mcode = langdict[None]
    while len(subtags) > 0:
        subtag = subtags.pop(0)
        if subtag not in langdict:
            subtag = subtag.title()
        if subtag not in langdict:
            subtag = subtag.upper()
        if subtag in langdict:
            mcode = langdict[subtag]
            break
    return pack('>HBB', 0, mcode[1], mcode[0])


def mobi2iana(langcode, sublangcode):
    prefix = suffix = None
    for code, d in IANA_MOBI.items():
        for subcode, t in d.items():
            cc, cl = t
            if cc == langcode:
                prefix = code
            if cl == sublangcode:
                suffix = subcode.lower() if subcode else None
                break
        if prefix is not None:
            break
    if prefix is None:
        return 'und'
    if suffix is None:
        return prefix
    return prefix + '-' + suffix



#!/usr/bin/env python
# License: GPL v3 Copyright: 2021, Kovid Goyal <kovid at kovidgoyal.net>

import importlib
import os
from pprint import pprint

QT_WRAPPER = 'PyQt6'

base = os.path.dirname(os.path.abspath(__file__))
module_lists = {
    'core': (
        'QtCore',
        'QtGui',
        'QtWidgets',
        'QtNetwork',
        'QtSvg',
        'QtPrintSupport',
        'QtOpenGL',
        'QtOpenGLWidgets',
        'QtQuick',
        'QtMultimedia',
        'QtMultimediaWidgets',
        'QtTextToSpeech',
    ),
    'webengine': (
        'QtWebEngineCore',
        'QtWebEngineWidgets',
        'QtWebChannel',
    ),
    'dbus': ('QtDBus',),
}


special_case = {
    'pyqtProperty': '''

_T = TypeVar('_T')
_R = TypeVar('_R')

class pyqtProperty:
    # --- Overload 1: Used as a decorator factory (e.g., @pyqtProperty(int)) ---
    @overload
    def __init__(
        self,
        type: type[_R] | str,
        fget: None = None,
        fset: None = None,
        freset: None = None,
        fdel: None = None,
        doc: str | None = None,
        designable: bool = True,
        scriptable: bool = True,
        stored: bool = True,
        user: bool = False,
        constant: bool = False,
        final: bool = False,
        notify: pyqtSignal | None = None,
        revision: int = 0,
    ) -> None: ...

    # --- Overload 2: Used as a functional constructor (e.g., x = pyqtProperty(int, get_x, set_x)) ---
    @overload
    def __init__(
        self,
        type: type[_R] | str,
        fget: Callable[[Any], _R],
        fset: Callable[[Any, _R], None] | None = None,
        freset: Callable[[Any], None] | None = None,
        fdel: Callable[[Any], None] | None = None,
        doc: str | None = None,
        designable: bool = True,
        scriptable: bool = True,
        stored: bool = True,
        user: bool = False,
        constant: bool = False,
        final: bool = False,
        notify: pyqtSignal | None = None,
        revision: int = 0,
    ) -> None: ...

    # --- Behaviors required to act like a standard Python property descriptor ---
    def __call__(self, fget: Callable[[_T], _R]) -> pyqtProperty: ...
    def __get__(self, instance: _T, owner: type[_T] | None = None) -> _R: ...
    def __set__(self, instance: _T, value: _R) -> None: ...
    def __delete__(self, instance: _T) -> None: ...

    def setter(self, fset: Callable[[_T, _R], None]) -> pyqtProperty: ...
    def deleter(self, fdel: Callable[[_T], None]) -> pyqtProperty: ...
''',
    # The following ty thinks dont exist probably cause they are missing from stub
    # files
    'QtAudio': '',
    'QStyleOptionMenuItemV2': '',
    'QPyAbstractRange': '',
    'QPySequenceRange': '',
    'QPyTableRange': '',
    'QRangeModel': '',
}


def scan(name):
    module_names = module_lists[name]
    name_map = {}
    types = []
    blocks = []
    for mod_name in module_names:
        mod = importlib.import_module(f'{QT_WRAPPER}.{mod_name}')
        full_name = name_map[mod_name] = mod.__name__
        types.append(f'import {full_name}')
        for obj_name in sorted(dir(mod)):
            if not obj_name.startswith('_') and obj_name not in name_map:
                name_map[obj_name] = full_name
                if (block := special_case.get(obj_name)) is not None:
                    blocks.append(block)
                else:
                    types.append(f'{obj_name} = {full_name}.{obj_name}')
    with open(f'{base}/{name}.py', 'w') as f:
        print('# autogenerated by __main__.py do not edit', file=f)
        print(f'from .{name}_name_map import module_names, name_map', file=f)
        print(
            '''from .loader import dynamic_load

already_imported = {}
qt_modules = {}

def __getattr__(name):
    return dynamic_load(name, name_map, already_imported, qt_modules, module_names)
''',
            end='',
            file=f,
        )
    if name == 'core':
        module_names += ('sip',)
        mod = importlib.import_module(f'{QT_WRAPPER}.sip')
        name_map['sip'] = mod.__name__
        types.append(f'import {mod.__name__}')
        types.append(f'sip = {mod.__name__}')

    with open(f'{base}/{name}.pyi', 'w') as f:
        print('# autogenerated by __main__.py do not edit', file=f)
        if name == 'core':
            print('from collections.abc import Callable', file=f)
            print('from typing import Any, TypeVar, overload', file=f)
        print('\n'.join(types), file=f)
        for block in blocks:
            if block:
                print(file=f)
                print(block.strip(), file=f)
    with open(f'{base}/{name}_name_map.py', 'w') as f:
        print('# autogenerated by __main__.py do not edit', file=f)
        print('name_map =', end=' ', file=f)
        pprint(name_map, stream=f)
        print('module_names = frozenset(', end='', file=f)
        pprint(module_names, stream=f)
        print(')', file=f)


top_level_module_names = ()
for name, module_name in module_lists.items():
    top_level_module_names += module_name
    scan(name)
with open(f'{base}/__init__.py', 'w') as f:
    print('# autogenerated by __main__.py do not edit', file=f)
    print('top_level_module_names = ', end='', file=f)
    pprint(top_level_module_names, stream=f)
    print(
        f'''

def __getattr__(name):
    if name in top_level_module_names:
        import importlib
        return importlib.import_module('{QT_WRAPPER}.' + name)
    raise AttributeError(name)
''',
        end='',
        file=f,
    )



#!/usr/bin/env python
# License: GPLv3 Copyright: 2009, Kovid Goyal <kovid@kovidgoyal.net>

import re
import time
from functools import partial
from typing import TYPE_CHECKING

from qt.core import QAction, QApplication, QComboBox, QCompleter, QDialog, QEvent, QIcon, QKeyEvent, QKeySequence, QLineEdit, Qt, QTimer, pyqtSignal, pyqtSlot

from calibre.gui2 import QT_HIDDEN_CLEAR_ACTION, config, gprefs, question_dialog
from calibre.gui2.dialogs.saved_search_editor import SavedSearchEditor
from calibre.gui2.dialogs.search import SearchDialog
from calibre.gui2.widgets import stylesheet_for_lineedit
from calibre.utils.icu import primary_sort_key
from calibre.utils.localization import _, pgettext

if TYPE_CHECKING:
    from calibre.gui2.ui import Main


class AsYouType(str):
    as_you_type: bool

    def __new__(cls, text):
        self = str.__new__(cls, text)
        self.as_you_type = True
        return self


class SearchLineEdit(QLineEdit):  # {{{
    key_pressed = pyqtSignal(object)
    clear_history = pyqtSignal()
    select_on_mouse_press = None
    as_url = None

    def keyPressEvent(self, a0):
        self.key_pressed.emit(a0)
        QLineEdit.keyPressEvent(self, a0)

    def dropEvent(self, a0):
        p = self.parent()
        assert p is not None
        assert isinstance(p, SearchBox2)
        p.normalize_state()
        return QLineEdit.dropEvent(self, a0)

    def contextMenuEvent(self, a0):
        p = self.parent()
        assert p is not None
        assert isinstance(p, SearchBox2)
        p.normalize_state()
        menu = self.createStandardContextMenu()
        assert menu is not None
        menu.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose)
        ac = menu.addAction(_('Paste and &search'))
        assert ac is not None
        clipboard = QApplication.clipboard()
        assert clipboard is not None
        ac.setEnabled(bool(clipboard.text()))
        ac.setIcon(QIcon.ic('search.png'))
        ac.triggered.connect(self.paste_and_search)
        for action in menu.actions():
            if action.text().startswith(_('&Paste') + '\t'):
                menu.insertAction(action, ac)
                break
        else:
            menu.addAction(ac)
        menu.addSeparator()
        ac = menu.addAction(_('Invert current search'))
        assert ac is not None
        ac.setEnabled(bool(self.text().strip()))
        ac.setIcon(QIcon.ic('search.png'))
        ac.triggered.connect(self.invert_search)
        menu.addAction(ac)
        menu.addSeparator()
        if self.as_url is not None:
            url = self.as_url(self.text())
            if url:

                def copy_as_url():
                    cb = QApplication.clipboard()
                    assert cb is not None
                    cb.setText(url)

                menu.addAction(_('Copy search as URL'), copy_as_url)
        clear_ac = menu.addAction(_('&Clear search history'))
        assert clear_ac is not None
        clear_ac.triggered.connect(self.clear_history)
        menu.exec(a0.globalPos())

    def invert_search(self):
        q = self.text().strip()
        if q:
            if q.startswith('NOT ( ') and q.endswith(' )'):
                q = q[6:-2]
            else:
                q = f'NOT ( {q} )'
            self.setText(q)
            ev = QKeyEvent(QEvent.Type.KeyPress, Qt.Key.Key_Enter, Qt.KeyboardModifier.NoModifier)
            self.keyPressEvent(ev)

    def paste_and_search(self):
        self.paste()
        ev = QKeyEvent(QEvent.Type.KeyPress, Qt.Key.Key_Enter, Qt.KeyboardModifier.NoModifier)
        self.keyPressEvent(ev)

    @pyqtSlot()
    def paste(self, *args):
        p = self.parent()
        assert p is not None
        assert isinstance(p, SearchBox2)
        p.normalize_state()
        return QLineEdit.paste(self)

    def focusInEvent(self, a0):
        self.select_on_mouse_press = time.time()
        return QLineEdit.focusInEvent(self, a0)

    def mousePressEvent(self, a0):
        QLineEdit.mousePressEvent(self, a0)
        if self.select_on_mouse_press is not None and abs(time.time() - self.select_on_mouse_press) < 0.2:
            self.selectAll()
        self.select_on_mouse_press = None


# }}}


class SearchBox2(QComboBox):  # {{{
    """
    To use this class:

        * Call initialize()
        * Connect to the search() and cleared() signals from this widget.
        * Connect to the changed() signal to know when the box content changes
        * Connect to focus_to_library() signal to be told to manually change focus
        * Call search_done() after every search is complete
        * Call set_search_string() to perform a search programmatically
        * You can use the current_text property to get the current search text
          Be aware that if you are using it in a slot connected to the
          changed() signal, if the connection is not queued it will not be
          accurate.
    """

    INTERVAL = 1500  #: Time to wait before emitting search signal
    MAX_COUNT = 25

    search = pyqtSignal(object)
    cleared = pyqtSignal()
    changed = pyqtSignal()
    focus_to_library = pyqtSignal()

    def __init__(self, parent=None, add_clear_action=True, as_url=None):
        QComboBox.__init__(self, parent)
        self.line_edit = SearchLineEdit(self)
        self.line_edit.as_url = as_url
        self.setLineEdit(self.line_edit)
        self.line_edit.clear_history.connect(self.clear_history)
        if add_clear_action:
            le = self.lineEdit()
            assert le is not None
            le.setClearButtonEnabled(True)
            ac = self.findChild(QAction, QT_HIDDEN_CLEAR_ACTION)
            if ac is not None:
                ac.triggered.connect(self.clear_clicked)

        c = self.line_edit.completer()
        assert c is not None
        c.setCompletionMode(QCompleter.CompletionMode.PopupCompletion)
        c.highlighted[str].connect(self.completer_used)

        self.line_edit.key_pressed.connect(self.key_pressed, type=Qt.ConnectionType.DirectConnection)
        # QueuedConnection as workaround for https://bugreports.qt-project.org/browse/QTBUG-40807
        self.textActivated.connect(self.history_selected, type=Qt.ConnectionType.QueuedConnection)
        self.setEditable(True)
        self.as_you_type = True
        self.timer = QTimer()
        self.timer.setSingleShot(True)
        self.timer.timeout.connect(self.timer_event, type=Qt.ConnectionType.QueuedConnection)
        self.setInsertPolicy(QComboBox.InsertPolicy.NoInsert)
        self.setMaxCount(self.MAX_COUNT)
        self.setSizeAdjustPolicy(QComboBox.SizeAdjustPolicy.AdjustToMinimumContentsLengthWithIcon)
        self.setMinimumContentsLength(25)
        self._in_a_search = False
        self.tool_tip_text = self.toolTip()
        self.parse_error_action = None

    def add_action(self, icon, position=QLineEdit.ActionPosition.TrailingPosition):
        if not isinstance(icon, QIcon):
            icon = QIcon.ic(icon)
        le = self.lineEdit()
        assert le is not None
        return le.addAction(icon, position)

    def initialize(self, opt_name, colorize=False, help_text=_('Search'), as_you_type=None):
        self.as_you_type = config['search_as_you_type'] if as_you_type is None else as_you_type
        self.opt_name = opt_name
        items = []
        for item in config[opt_name]:
            if item not in items:
                items.append(item)
        self.addItems(items)
        self.line_edit.setPlaceholderText(help_text)
        self.clear()

    def clear_history(self):
        config[self.opt_name] = []
        super().clear()
        self.clear()

    clear_search_history = clear_history

    def hide_completer_popup(self):
        try:
            le = self.lineEdit()
            assert le is not None
            comp = le.completer()
            assert comp is not None
            pop = comp.popup()
            assert pop is not None
            pop.setVisible(False)
        except Exception:
            pass

    def normalize_state(self):
        self.setToolTip(self.tool_tip_text)
        self.setStyleSheet('')
        self.show_parse_error_action(False)

    def text(self):
        return self.currentText()

    def clear(self, emit_search=True):
        self.show_parse_error_action(False)
        self.normalize_state()
        self.setEditText('')
        if emit_search:
            self.search.emit('')
        self._in_a_search = False
        self.cleared.emit()

    def clear_clicked(self, *args):
        self.clear()
        self.setFocus(Qt.FocusReason.OtherFocusReason)

    def show_parse_error_action(self, to_show, tooltip=''):
        if self.parse_error_action is not None:
            self.parse_error_action.setVisible(to_show)
            self.parse_error_action.setToolTip(tooltip)

    def search_done(self, ok):
        if isinstance(ok, (str, bytes)):
            self.setToolTip(ok)
            self.show_parse_error_action(True, tooltip=ok)
            ok = False
        if not str(self.currentText()).strip():
            self.setStyleSheet('')
            self.clear(emit_search=False)
            return
        self._in_a_search = ok
        if self.parse_error_action is not None and not ok:
            self.setStyleSheet(stylesheet_for_lineedit(bool(ok), 'QComboBox'))
        else:
            self.setStyleSheet('')

    # Comes from the lineEdit control
    def key_pressed(self, event):
        k = event.key()
        if k in (
            Qt.Key.Key_Left,
            Qt.Key.Key_Right,
            Qt.Key.Key_Up,
            Qt.Key.Key_Down,
            Qt.Key.Key_Home,
            Qt.Key.Key_End,
            Qt.Key.Key_PageUp,
            Qt.Key.Key_PageDown,
            Qt.Key.Key_unknown,
        ):
            return
        self.normalize_state()
        if self._in_a_search:
            self.changed.emit()
            self._in_a_search = False
        if event.key() in (Qt.Key.Key_Return, Qt.Key.Key_Enter):
            self.do_search()
            self.focus_to_library.emit()
        elif self.as_you_type and str(event.text()):
            self.timer.start(1500)

    # Comes from the combobox itself
    def keyPressEvent(self, e):
        self.show_parse_error_action(False)
        k = e.key()
        if k in (Qt.Key.Key_Enter, Qt.Key.Key_Return):
            return self.do_search()
        if k not in (Qt.Key.Key_Up, Qt.Key.Key_Down):
            return QComboBox.keyPressEvent(self, e)
        self.blockSignals(True)
        self.normalize_state()
        le = self.lineEdit()
        assert le is not None
        if k == Qt.Key.Key_Down and self.currentIndex() == 0 and not le.text():
            self.setCurrentIndex(1), self.setCurrentIndex(0)
            e.accept()
        else:
            QComboBox.keyPressEvent(self, e)
        self.blockSignals(False)

    def completer_used(self, text):
        self.timer.stop()
        self.normalize_state()

    def timer_event(self):
        self._do_search(as_you_type=True)
        # since this is an automatic search keep focus
        self.setFocus(Qt.FocusReason.OtherFocusReason)

    def history_selected(self, text):
        self.changed.emit()
        self.do_search()

    def _do_search(self, store_in_history=True, as_you_type=False):
        self.hide_completer_popup()
        text = str(self.currentText()).strip()
        if not text:
            return self.clear()
        if as_you_type:
            text = AsYouType(text)
        self.search.emit(text)

        if store_in_history:
            idx = self.findText(text, Qt.MatchFlag.MatchFixedString | Qt.MatchFlag.MatchCaseSensitive)
            self.block_signals(True)
            if idx < 0:
                self.insertItem(0, text)
            else:
                t = self.itemText(idx)
                self.removeItem(idx)
                self.insertItem(0, t)
            self.setCurrentIndex(0)
            self.block_signals(False)
            history = [str(self.itemText(i)) for i in range(self.count())]
            config[self.opt_name] = history

    def do_search(self, *args):
        self._do_search()
        self.timer.stop()

    def block_signals(self, yes):
        self.blockSignals(yes)
        self.line_edit.blockSignals(yes)

    def set_search_string(self, txt, store_in_history=False, emit_changed=True):
        if not store_in_history:
            self.textActivated.disconnect()
        try:
            self.setFocus(Qt.FocusReason.OtherFocusReason)
            if not txt:
                self.clear()
            else:
                self.normalize_state()
                # must turn on case sensitivity here so that tag browser strings
                # are not case-insensitively replaced from history
                line_edit_completer = self.line_edit.completer()
                assert line_edit_completer is not None
                line_edit_completer.setCaseSensitivity(Qt.CaseSensitivity.CaseSensitive)
                self.setEditText(txt)
                self.line_edit.end(False)
                if emit_changed:
                    self.changed.emit()
                self._do_search(store_in_history=store_in_history)
                line_edit_completer.setCaseSensitivity(Qt.CaseSensitivity.CaseInsensitive)
            self.focus_to_library.emit()
        finally:
            if not store_in_history:
                # QueuedConnection as workaround for https://bugreports.qt-project.org/browse/QTBUG-40807
                self.textActivated.connect(self.history_selected, type=Qt.ConnectionType.QueuedConnection)

    def search_as_you_type(self, enabled):
        self.as_you_type = enabled

    def in_a_search(self):
        return self._in_a_search

    @property
    def current_text(self):
        le = self.lineEdit()
        assert le is not None
        return str(le.text())

    # }}}


class SearchBoxMixin:  # {{{
    def init_search_box_mixin(self: Main):
        self.search.initialize('main_search_history', help_text=_('Search (For advanced search click the gear icon to the left)'))
        self.search.cleared.connect(self.search_box_cleared)
        # Queued so that search.current_text will be correct
        self.search.changed.connect(self.search_box_changed, type=Qt.ConnectionType.QueuedConnection)
        self.search.focus_to_library.connect(self.focus_to_library)
        self.advanced_search_toggle_action.triggered.connect(self.do_advanced_search)

        self.search.clear()
        self.search.setMaximumWidth(self.width() - 150)
        self.action_focus_search = QAction(self)
        shortcuts = [str(x.toString(QKeySequence.SequenceFormat.PortableText)) for x in QKeySequence.keyBindings(QKeySequence.StandardKey.Find)]
        shortcuts += ['/', 'Alt+S']
        self.keyboard.register_shortcut('start search', _('Start search'), default_keys=shortcuts, action=self.action_focus_search)
        self.action_focus_search.triggered.connect(self.focus_search_box)
        self.addAction(self.action_focus_search)
        self.search.setStatusTip(re.sub(r'<\w+>', ' ', str(self.search.toolTip())))
        self.set_highlight_only_button_icon()
        self.highlight_only_button.clicked.connect(self.highlight_only_clicked)
        tt = _('Enable or disable search highlighting.') + '<br><br>'
        tt += config.help('highlight_search_matches')
        self.highlight_only_button.setToolTip(tt)
        self.highlight_only_action = ac = QAction(self)
        self.addAction(ac), ac.triggered.connect(self.highlight_only_clicked)
        self.keyboard.register_shortcut('highlight search results', _('Highlight search results'), action=self.highlight_only_action)
        self.refresh_search_bar_widgets()

    def refresh_search_bar_widgets(self: Main):
        self.set_highlight_only_button_icon()
        if gprefs['search_tool_bar_shows_text']:
            self.search_bar.search_button.setText(_('Search'))
            self.search_bar.search_button.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonTextBesideIcon)
        else:
            self.search_bar.search_button.setText(None)
            self.search_bar.search_button.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonIconOnly)

    def highlight_only_clicked(self: Main, state):
        if not config['highlight_search_matches'] and not question_dialog(
            self,
            _('Are you sure?'),
            _(
                'This will change how searching works. When you search, instead of showing only the '
                'matching books, all books will be shown with the matching books highlighted. '
                'Are you sure this is what you want?'
            ),
            skip_dialog_name='confirm_search_highlight_toggle',
        ):
            return
        config['highlight_search_matches'] = not config['highlight_search_matches']
        self.set_highlight_only_button_icon()
        self.search.do_search()
        self.focus_to_library()

    def set_highlight_only_button_icon(self: Main):
        b = self.highlight_only_button
        if config['highlight_search_matches']:
            b.setIcon(QIcon.ic('highlight_only_on.png'))
            if gprefs['search_tool_bar_shows_text']:
                b.setText(_('Filter'))
                b.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonTextBesideIcon)
            else:
                b.setText(None)
                b.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonIconOnly)
        else:
            b.setIcon(QIcon.ic('highlight_only_off.png'))
            if gprefs['search_tool_bar_shows_text']:
                b.setText(pgettext('mark books matching search result instead of filtering them', 'Highlight'))
                b.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonTextBesideIcon)
            else:
                b.setText(None)
                b.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonIconOnly)
        self.highlight_only_button.setVisible(gprefs['show_highlight_toggle_button'])
        self.library_view._model.set_highlight_only(config['highlight_search_matches'])

    def focus_search_box(self: Main, *args):
        self.search.setFocus(Qt.FocusReason.OtherFocusReason)
        le = self.search.lineEdit()
        if le:
            le.selectAll()

    def search_box_cleared(self: Main):
        self.tags_view.clear()
        self.set_number_of_books_shown()

    def search_box_changed(self: Main):
        self.tags_view.conditional_clear(self.search.current_text)

    def do_advanced_search(self: Main, *args):
        d = SearchDialog(self, self.current_db)
        if d.exec() == QDialog.DialogCode.Accepted:
            self.search.set_search_string(d.search_string(), store_in_history=True)

    def do_search_button(self: Main):
        self.search.do_search()
        self.focus_to_library()

    def focus_to_library(self: Main):
        self.focus_current_view()

    # }}}


class SavedSearchBoxMixin:  # {{{
    def init_saved_seach_box_mixin(self: Main):
        pass

    def add_saved_searches_to_menu(self: Main, menu, db, add_action_func=None):
        def add_action(current_menu, whole_name, last_component, func=None):
            if add_action_func is None:
                return current_menu.addAction(last_component, func)
            return add_action_func(current_menu, whole_name, last_component)

        folder_icon = QIcon.ic('folder_saved_search.png')
        search_icon = QIcon.ic('search.png')
        use_hierarchy = 'search' in db.new_api.pref('categories_using_hierarchy', [])
        submenus = {}
        for name in sorted(db.saved_search_names(), key=lambda x: primary_sort_key(x.strip())):
            display_name = name.replace('&', '&&')
            current_menu = menu
            if use_hierarchy:
                components = tuple(n.strip() for n in name.split('.'))
                hierarchy = components[:-1]
                last = components[-1]
                # Walk the hierarchy, creating submenus as needed
                for i, c in enumerate(hierarchy, start=1):
                    hierarchical_prefix = '.'.join(hierarchy[:i])
                    if hierarchical_prefix not in submenus:
                        current_menu = current_menu.addMenu(c.replace('&', '&&'))
                        current_menu.setIcon(folder_icon)
                        submenus[hierarchical_prefix] = current_menu
                    else:
                        current_menu = submenus[hierarchical_prefix]
                ac = add_action(
                    current_menu,
                    display_name,
                    last.replace('&', '&&'),
                    partial(self.search.set_search_string, 'search:"=' + name + '"'),
                )
            else:
                ac = add_action(
                    current_menu,
                    display_name,
                    display_name,
                    partial(self.search.set_search_string, 'search:"=' + name + '"'),
                )
            if ac.icon().isNull():
                ac.setIcon(search_icon)

    def populate_add_saved_search_menu(self: Main, to_menu):
        m = to_menu
        m.clear()
        m.addAction(QIcon.ic('search_add_saved.png'), _('Add Saved search'), self.add_saved_search)
        m.addAction(QIcon.ic('search_copy_saved.png'), _('Get Saved search expression'), self.get_saved_search_text)
        m.addAction(QIcon.ic('folder_saved_search.png'), _('Manage Saved searches'), partial(self.do_saved_search_edit, None))
        m.addSeparator()
        self.add_saved_searches_to_menu(m, self.current_db)

    def saved_searches_changed(self: Main, set_restriction=None, recount=True):
        self.build_search_restriction_list()
        if recount:
            self.tags_view.recount()
        if set_restriction:  # redo the search restriction if there was one
            self.apply_named_search_restriction(set_restriction)

    def do_saved_search_edit(self: Main, search):
        d = SavedSearchEditor(self, search)
        d.exec()
        if d.result() == QDialog.DialogCode.Accepted:
            self.do_rebuild_saved_searches()

    def do_rebuild_saved_searches(self: Main):
        self.saved_searches_changed()

    def add_saved_search(self: Main):
        from calibre.gui2.dialogs.saved_search_editor import AddSavedSearch

        d = AddSavedSearch(parent=self, search=self.search.current_text)
        if d.exec() == QDialog.DialogCode.Accepted:
            self.current_db.new_api.ensure_has_search_category(fail_on_existing=False)
            self.do_rebuild_saved_searches()

    def get_saved_search_text(self: Main, search_name=None):
        db = self.current_db
        try:
            current_search = search_name or self.search.currentText()
            if not current_search.startswith('search:'):
                raise ValueError()
            # This strange expression accounts for the four ways a search can be written:
            # search:fff, search:"fff", search:"=fff". and search:="fff"
            current_search = current_search[7:].lstrip('=').strip('"').lstrip('=')
            current_search = db.saved_search_lookup(current_search)
            if not current_search:
                raise ValueError()
            self.search.set_search_string(current_search)
        except Exception:
            from calibre.gui2.ui import get_gui

            get_gui(fail_if_absent=True).status_bar.show_message(_('Current search is not a saved search'), 3000)

    # }}}



#!/usr/bin/env python
# License: GPLv3 Copyright: 2013, Kovid Goyal <kovid at kovidgoyal.net>

from calibre.customize.conversion import OptionRecommendation, OutputFormatPlugin
from calibre.utils.localization import _

PAGE_SIZES = ['a0', 'a1', 'a2', 'a3', 'a4', 'a5', 'a6', 'b0', 'b1', 'b2', 'b3', 'b4', 'b5', 'b6', 'legal', 'letter']


class DOCXOutput(OutputFormatPlugin):
    name = 'DOCX Output'
    author = 'Kovid Goyal'
    file_type = 'docx'
    commit_name = 'docx_output'
    ui_data = {'page_sizes': PAGE_SIZES}

    options = {
        OptionRecommendation(
            name='docx_page_size',
            recommended_value='letter',
            level=OptionRecommendation.LOW,
            choices=PAGE_SIZES,
            help=_('The size of the page. Default is letter. Choices are %s') % PAGE_SIZES,
        ),
        OptionRecommendation(
            name='docx_custom_page_size',
            recommended_value=None,
            help=_(
                'Custom size of the document. Use the form width x height, '
                'for example: `123x321` to specify the width and height (in pts). '
                'This overrides any specified page-size.'
            ),
        ),
        OptionRecommendation(
            name='docx_no_cover',
            recommended_value=False,
            help=_('Do not insert the book cover as an image at the start of the document. If you use this option, the book cover will be discarded.'),
        ),
        OptionRecommendation(
            name='preserve_cover_aspect_ratio',
            recommended_value=False,
            help=_('Preserve the aspect ratio of the cover image instead of stretching it out to cover the entire page.'),
        ),
        OptionRecommendation(
            name='docx_no_toc',
            recommended_value=False,
            help=_('Do not insert the table of contents as a page at the start of the document.'),
        ),
        OptionRecommendation(
            name='extract_to',
            help=_('Extract the contents of the generated %s file to the specified folder. The contents of the folder are first deleted, so be careful.')
            % 'DOCX',
        ),
        OptionRecommendation(
            name='docx_page_margin_left',
            recommended_value=72.0,
            level=OptionRecommendation.LOW,
            help=_('The size of the left page margin, in pts. Default is 72pt. Overrides the common left page margin setting.'),
        ),
        OptionRecommendation(
            name='docx_page_margin_top',
            recommended_value=72.0,
            level=OptionRecommendation.LOW,
            help=_('The size of the top page margin, in pts. Default is 72pt. Overrides the common top page margin setting, unless set to zero.'),
        ),
        OptionRecommendation(
            name='docx_page_margin_right',
            recommended_value=72.0,
            level=OptionRecommendation.LOW,
            help=_('The size of the right page margin, in pts. Default is 72pt. Overrides the common right page margin setting, unless set to zero.'),
        ),
        OptionRecommendation(
            name='docx_page_margin_bottom',
            recommended_value=72.0,
            level=OptionRecommendation.LOW,
            help=_('The size of the bottom page margin, in pts. Default is 72pt. Overrides the common bottom page margin setting, unless set to zero.'),
        ),
    }

    def convert_metadata(self, oeb):
        from io import BytesIO

        from lxml import etree

        from calibre.ebooks.metadata.opf2 import OPF as ReadOPF
        from calibre.ebooks.oeb.base import OPF, OPF2_NS

        package = etree.Element(OPF('package'), attrib={'version': '2.0'}, nsmap={None: OPF2_NS})
        oeb.metadata.to_opf2(package)
        self.mi = ReadOPF(BytesIO(etree.tostring(package, encoding='utf-8')), populate_spine=False, try_to_guess_cover=False).to_book_metadata()

    def convert(self, oeb_book, output, input_plugin, opts, log):
        oeb = oeb_book
        output_path = output
        from calibre.ebooks.docx.writer.container import DOCX
        from calibre.ebooks.docx.writer.from_html import Convert

        docx = DOCX(opts, log)
        self.convert_metadata(oeb)
        Convert(oeb, docx, self.mi, not opts.docx_no_cover, not opts.docx_no_toc)()
        docx.write(output_path, self.mi)
        if opts.extract_to:
            from calibre.ebooks.docx.dump import do_dump

            do_dump(output_path, opts.extract_to)



# License: GPLv3 Copyright: 2009, Kovid Goyal <kovid@kovidgoyal.net>

import os

from calibre.customize.conversion import InputFormatPlugin
from calibre.utils.localization import _


class MOBIInput(InputFormatPlugin):
    name = 'MOBI Input'
    author = 'Kovid Goyal'
    description = _('Convert MOBI files (.mobi, .prc, .azw) to HTML')
    file_types = {'mobi', 'prc', 'azw', 'azw3', 'pobi'}
    commit_name = 'mobi_input'

    def convert(self, stream, options, file_ext, log, accelerators):
        self.is_kf8 = False
        self.mobi_is_joint = False

        from lxml import html

        from calibre.ebooks.mobi.reader.mobi6 import MobiReader

        parse_cache = {}
        try:
            mr = MobiReader(stream, log, options.input_encoding, options.debug_pipeline)
            if mr.kf8_type is None:
                mr.extract_content('.', parse_cache)

        except Exception:
            mr = MobiReader(stream, log, options.input_encoding, options.debug_pipeline, try_extra_data_fix=True)
            if mr.kf8_type is None:
                mr.extract_content('.', parse_cache)

        if mr.kf8_type is not None:
            log(f'Found KF8 MOBI of type {mr.kf8_type!r}')
            if mr.kf8_type == 'joint':
                self.mobi_is_joint = True
            from calibre.ebooks.mobi.reader.mobi8 import Mobi8Reader

            mr = Mobi8Reader(mr, log)
            opf = os.path.abspath(mr())
            self.encrypted_fonts = mr.encrypted_fonts
            self.is_kf8 = True
            return opf

        raw = parse_cache.pop('calibre_raw_mobi_markup', False)
        if raw:
            if isinstance(raw, str):
                raw = raw.encode('utf-8')
            with open('debug-raw.html', 'wb') as f:
                f.write(raw)
        from calibre.ebooks.oeb.base import close_self_closing_tags

        for f, root in parse_cache.items():
            raw = html.tostring(root, encoding='utf-8', method='xml', include_meta_content_type=False)
            raw = close_self_closing_tags(raw)
            with open(f, 'wb') as q:
                q.write(raw)
        accelerators['pagebreaks'] = '//h:div[@class="mbp_pagebreak"]'
        return mr.created_opf_path



# License: GPLv3 Copyright: 2009, Kovid Goyal <kovid@kovidgoyal.net>

from calibre.customize import Plugin as _Plugin
from calibre.utils.localization import _

FONT_SIZES = [
    ('xx-small', 1),
    ('x-small', None),
    ('small', 2),
    ('medium', 3),
    ('large', 4),
    ('x-large', 5),
    ('xx-large', 6),
    (None, 7),
]


class Plugin(_Plugin):
    fbase = 12
    font_sizes = (5, 7, 9, 12, 13.5, 17, 20, 22, 24)
    screen_size = (1600, 1200)
    dpi = 100

    def __init__(self, *args, **kwargs):
        _Plugin.__init__(self, *args, **kwargs)
        self.width, self.height = self.screen_size
        self.fkey = list(self.font_sizes)
        self.fsizes = []
        for (name, num), size in zip(FONT_SIZES, self.fkey):
            self.fsizes.append((name, num, float(size)))
        self.fnames = {name: sz for name, _, sz in self.fsizes if name}
        self.fnums = {num: sz for _, num, sz in self.fsizes if num}
        self.width_pts = self.width * 72.0 / self.dpi
        self.height_pts = self.height * 72.0 / self.dpi


# Input profiles {{{


class InputProfile(Plugin):
    author = 'Kovid Goyal'
    supported_platforms = {'windows', 'osx', 'linux'}
    can_be_disabled = False
    type = _('Input profile')

    name = 'Default Input Profile'
    short_name = 'default'  # Used in the CLI so don't use spaces etc. in it
    description = _('This profile tries to provide sane defaults and is useful if you know nothing about the input document.')


class SonyReaderInput(InputProfile):
    name = 'Sony Reader'
    short_name = 'sony'
    description = _('This profile is intended for the SONY PRS line. The 500/505/600/700 etc.')

    screen_size = (584, 754)
    dpi = 168.451
    fbase = 12
    font_sizes = (7.5, 9, 10, 12, 15.5, 20, 22, 24)


class SonyReader300Input(SonyReaderInput):
    name = 'Sony Reader 300'
    short_name = 'sony300'
    description = _('This profile is intended for the SONY PRS 300.')

    dpi = 200


class SonyReader900Input(SonyReaderInput):
    author = 'John Schember'
    name = 'Sony Reader 900'
    short_name = 'sony900'
    description = _('This profile is intended for the SONY PRS-900.')

    screen_size = (584, 978)


class MSReaderInput(InputProfile):
    name = 'Microsoft Reader'
    short_name = 'msreader'
    description = _('This profile is intended for the Microsoft Reader.')

    screen_size = (480, 652)
    dpi = 96
    fbase = 13
    font_sizes = (10, 11, 13, 16, 18, 20, 22, 26)


class MobipocketInput(InputProfile):
    name = 'Mobipocket Books'
    short_name = 'mobipocket'
    description = _('This profile is intended for the Mobipocket books.')

    # Unfortunately MOBI books are not narrowly targeted, so this information is
    # quite likely to be spurious
    screen_size = (600, 800)
    dpi = 96
    fbase = 18
    font_sizes = (14, 14, 16, 18, 20, 22, 24, 26)


class HanlinV3Input(InputProfile):
    name = 'Hanlin V3'
    short_name = 'hanlinv3'
    description = _('This profile is intended for the Hanlin V3 and its clones.')

    # Screen size is a best guess
    screen_size = (584, 754)
    dpi = 168.451
    fbase = 16
    font_sizes = (12, 12, 14, 16, 18, 20, 22, 24)


class HanlinV5Input(HanlinV3Input):
    name = 'Hanlin V5'
    short_name = 'hanlinv5'
    description = _('This profile is intended for the Hanlin V5 and its clones.')

    # Screen size is a best guess
    screen_size = (584, 754)
    dpi = 200


class CybookG3Input(InputProfile):
    name = 'Cybook G3'
    short_name = 'cybookg3'
    description = _('This profile is intended for the Cybook G3.')

    # Screen size is a best guess
    screen_size = (600, 800)
    dpi = 168.451
    fbase = 16
    font_sizes = (12, 12, 14, 16, 18, 20, 22, 24)


class CybookOpusInput(InputProfile):
    author = 'John Schember'
    name = 'Cybook Opus'
    short_name = 'cybook_opus'
    description = _('This profile is intended for the Cybook Opus.')

    # Screen size is a best guess
    screen_size = (600, 800)
    dpi = 200
    fbase = 16
    font_sizes = (12, 12, 14, 16, 18, 20, 22, 24)


class KindleInput(InputProfile):
    name = 'Kindle'
    short_name = 'kindle'
    description = _('This profile is intended for the Amazon Kindle.')

    # Screen size is a best guess
    screen_size = (525, 640)
    dpi = 168.451
    fbase = 16
    font_sizes = (12, 12, 14, 16, 18, 20, 22, 24)


class IlliadInput(InputProfile):
    name = 'Illiad'
    short_name = 'illiad'
    description = _('This profile is intended for the Irex Illiad.')

    screen_size = (760, 925)
    dpi = 160.0
    fbase = 12
    font_sizes = (7.5, 9, 10, 12, 15.5, 20, 22, 24)


class IRexDR1000Input(InputProfile):
    author = 'John Schember'
    name = 'IRex Digital Reader 1000'
    short_name = 'irexdr1000'
    description = _('This profile is intended for the IRex Digital Reader 1000.')

    # Screen size is a best guess
    screen_size = (1024, 1280)
    dpi = 160
    fbase = 16
    font_sizes = (12, 14, 16, 18, 20, 22, 24)


class IRexDR800Input(InputProfile):
    author = 'Eric Cronin'
    name = 'IRex Digital Reader 800'
    short_name = 'irexdr800'
    description = _('This profile is intended for the IRex Digital Reader 800.')

    screen_size = (768, 1024)
    dpi = 160
    fbase = 16
    font_sizes = (12, 14, 16, 18, 20, 22, 24)


class NookInput(InputProfile):
    author = 'John Schember'
    name = 'Nook'
    short_name = 'nook'
    description = _('This profile is intended for the B&N Nook.')

    # Screen size is a best guess
    screen_size = (600, 800)
    dpi = 167
    fbase = 16
    font_sizes = (12, 12, 14, 16, 18, 20, 22, 24)


input_profiles = [
    InputProfile,
    SonyReaderInput,
    SonyReader300Input,
    SonyReader900Input,
    MSReaderInput,
    MobipocketInput,
    HanlinV3Input,
    HanlinV5Input,
    CybookG3Input,
    CybookOpusInput,
    KindleInput,
    IlliadInput,
    IRexDR1000Input,
    IRexDR800Input,
    NookInput,
]

input_profiles.sort(key=lambda x: x.name.lower())

# }}}


class OutputProfile(Plugin):
    author = 'Kovid Goyal'
    supported_platforms = {'windows', 'osx', 'linux'}
    can_be_disabled = False
    type = _('Output profile')

    name = 'Default Output Profile'
    short_name = 'default'  # Used in the CLI so don't use spaces etc. in it
    description = _(
        'This profile tries to provide sane defaults and is useful '
        'if you want to produce a document intended to be read at a '
        'computer or on a range of devices.'
    )

    #: The image size for comics
    comic_screen_size = (584, 754)

    #: If True the MOBI renderer on the device supports MOBI indexing
    supports_mobi_indexing = False

    #: If True output should be optimized for a touchscreen interface
    touchscreen = False
    touchscreen_news_css = ''
    #: A list of extra (beyond CSS 2.1) modules supported by the device
    #: Format is a css_parser profile dictionary (see iPad for example)
    extra_css_modules = []
    #: If True, the date is appended to the title of downloaded news
    periodical_date_in_title = True

    #: Characters used in jackets and catalogs
    ratings_char = '*'
    empty_ratings_char = ' '

    #: Unsupported unicode characters to be replaced during preprocessing
    unsupported_unicode_chars = []

    #: Number of ems that the left margin of a blockquote is rendered as
    mobi_ems_per_blockquote = 1.0

    #: Special periodical formatting needed in EPUB
    epub_periodical_format = None


class iPadOutput(OutputProfile):
    name = 'iPad'
    short_name = 'ipad'
    description = _('Intended for the iPad and similar devices with a resolution of 768x1024')
    screen_size = (768, 1024)
    comic_screen_size = (768, 1024)
    dpi = 132.0
    extra_css_modules = [
        {
            'name': 'webkit',
            'props': {
                '-webkit-border-bottom-left-radius': '{length}',
                '-webkit-border-bottom-right-radius': '{length}',
                '-webkit-border-top-left-radius': '{length}',
                '-webkit-border-top-right-radius': '{length}',
                '-webkit-border-radius': r'{border-width}(\s+{border-width}){0,3}|inherit',
            },
            'macros': {'border-width': '{length}|medium|thick|thin'},
        }
    ]

    ratings_char = '★'  # filled star
    empty_ratings_char = '☆'  # hollow star

    touchscreen = True
    # touchscreen_news_css {{{
    touchscreen_news_css = '''
            /* hr used in articles */
            .article_articles_list {
                width:18%;
                }
            .article_link {
                color: #593f29;
                font-style: italic;
                }
            .article_next {
                -webkit-border-top-right-radius:4px;
                -webkit-border-bottom-right-radius:4px;
                font-style: italic;
                width:32%;
                }

            .article_prev {
                -webkit-border-top-left-radius:4px;
                -webkit-border-bottom-left-radius:4px;
                font-style: italic;
                width:32%;
                }
            .article_sections_list {
                width:18%;
                }
            .articles_link {
                font-weight: bold;
                }
            .sections_link {
                font-weight: bold;
                }

            .caption_divider {
                border:#ccc 1px solid;
                }

            .touchscreen_navbar {
                background:#c3bab2;
                border:#ccc 0px solid;
                border-collapse:separate;
                border-spacing:1px;
                margin-left: 5%;
                margin-right: 5%;
                page-break-inside:avoid;
                width: 90%;
                -webkit-border-radius:4px;
                }
            .touchscreen_navbar td {
                background:#fff;
                font-family:Helvetica;
                font-size:80%;
                /* UI touchboxes use 8px padding */
                padding: 6px;
                text-align:center;
                }

            .touchscreen_navbar td a:link {
                color: #593f29;
                text-decoration: none;
                }

            /* Index formatting */
            .publish_date {
                text-align:center;
                }
            .divider {
                border-bottom:1em solid white;
                border-top:1px solid gray;
                }

            hr.caption_divider {
                border-color:black;
                border-style:solid;
                border-width:1px;
                }

            /* Feed summary formatting */
            .article_summary {
                display:inline-block;
                padding-bottom:0.5em;
                }
            .feed {
                font-family:sans-serif;
                font-weight:bold;
                font-size:larger;
                }

            .feed_link {
                font-style: italic;
                }

            .feed_next {
                -webkit-border-top-right-radius:4px;
                -webkit-border-bottom-right-radius:4px;
                font-style: italic;
                width:40%;
                }

            .feed_prev {
                -webkit-border-top-left-radius:4px;
                -webkit-border-bottom-left-radius:4px;
                font-style: italic;
                width:40%;
                }

            .feed_title {
                text-align: center;
                font-size: 160%;
                }

            .feed_up {
                font-weight: bold;
                width:20%;
                }

            .summary_headline {
                font-weight:bold;
                text-align:left;
                }

            .summary_byline {
                text-align:left;
                font-family:monospace;
                }

            .summary_text {
                text-align:left;
                }

        '''
    # }}}


class iPad3Output(iPadOutput):
    screen_size = comic_screen_size = (2048, 1536)
    dpi = 264.0
    name = 'iPad 3'
    short_name = 'ipad3'
    description = _('Intended for the iPad 3 and similar devices with a resolution of 1536x2048')


class TabletOutput(iPadOutput):
    name = _('Tablet')
    short_name = 'tablet'
    description = _('Intended for generic tablet devices, does no resizing of images')

    screen_size = (10000, 10000)
    comic_screen_size = screen_size


class SamsungGalaxy(TabletOutput):
    name = 'Samsung Galaxy'
    short_name = 'galaxy'
    description = _('Intended for the Samsung Galaxy and similar tablet devices with a resolution of 600x1280')
    screen_size = comic_screen_size = (600, 1280)


class NookHD(TabletOutput):
    name = 'Nook HD+'
    short_name = 'nook_hd_plus'
    description = _('Intended for the Nook HD+ and similar tablet devices with a resolution of 1280x1920')
    screen_size = comic_screen_size = (1280, 1920)


class SonyReaderOutput(OutputProfile):
    name = 'Sony Reader'
    short_name = 'sony'
    description = _('This profile is intended for the SONY PRS line. The 500/505/600/700 etc.')

    screen_size = (590, 775)
    dpi = 168.451
    fbase = 12
    font_sizes = (7.5, 9, 10, 12, 15.5, 20, 22, 24)
    unsupported_unicode_chars = ['\u201f', '\u201b']

    epub_periodical_format = 'sony'
    # periodical_date_in_title = False


class KoboReaderOutput(OutputProfile):
    name = 'Kobo Reader'
    short_name = 'kobo'

    description = _('This profile is intended for the Kobo Reader.')

    screen_size = (536, 710)
    comic_screen_size = (536, 710)
    dpi = 168.451
    fbase = 12
    font_sizes = (7.5, 9, 10, 12, 15.5, 20, 22, 24)


class SonyReader300Output(SonyReaderOutput):
    author = 'John Schember'
    name = 'Sony Reader 300'
    short_name = 'sony300'
    description = _('This profile is intended for the SONY PRS-300.')

    dpi = 200


class SonyReader900Output(SonyReaderOutput):
    author = 'John Schember'
    name = 'Sony Reader 900'
    short_name = 'sony900'
    description = _('This profile is intended for the SONY PRS-900.')

    screen_size = (600, 999)
    comic_screen_size = screen_size


class SonyReaderT3Output(SonyReaderOutput):
    author = 'Kovid Goyal'
    name = 'Sony Reader T3'
    short_name = 'sonyt3'
    description = _('This profile is intended for the SONY PRS-T3.')

    screen_size = (758, 934)
    comic_screen_size = screen_size


class GenericEink(SonyReaderOutput):
    name = _('Generic e-ink')
    short_name = 'generic_eink'
    description = _('Suitable for use with any e-ink device')
    epub_periodical_format = None


class GenericEinkLarge(GenericEink):
    name = _('Generic e-ink large')
    short_name = 'generic_eink_large'
    description = _('Suitable for use with any large screen e-ink device')

    screen_size = (600, 999)
    comic_screen_size = screen_size


class GenericEinkHD(GenericEink):
    name = _('Generic e-ink HD')
    short_name = 'generic_eink_hd'
    description = _('Suitable for use with any modern high resolution e-ink device')

    screen_size = (10000, 10000)
    comic_screen_size = screen_size


class JetBook5Output(OutputProfile):
    name = 'JetBook 5-inch'
    short_name = 'jetbook5'
    description = _('This profile is intended for the 5-inch JetBook.')

    screen_size = (480, 640)
    dpi = 168.451


class SonyReaderLandscapeOutput(SonyReaderOutput):
    name = 'Sony Reader Landscape'
    short_name = 'sony-landscape'
    description = _('This profile is intended for the SONY PRS line. The 500/505/700 etc, in landscape mode. Mainly useful for comics.')

    screen_size = (784, 1012)
    comic_screen_size = (784, 1012)


class MSReaderOutput(OutputProfile):
    name = 'Microsoft Reader'
    short_name = 'msreader'
    description = _('This profile is intended for the Microsoft Reader.')

    screen_size = (480, 652)
    dpi = 96
    fbase = 13
    font_sizes = (10, 11, 13, 16, 18, 20, 22, 26)


class MobipocketOutput(OutputProfile):
    name = 'Mobipocket Books'
    short_name = 'mobipocket'
    description = _('This profile is intended for the Mobipocket books.')

    # Unfortunately MOBI books are not narrowly targeted, so this information is
    # quite likely to be spurious
    screen_size = (600, 800)
    dpi = 96
    fbase = 18
    font_sizes = (14, 14, 16, 18, 20, 22, 24, 26)


class HanlinV3Output(OutputProfile):
    name = 'Hanlin V3'
    short_name = 'hanlinv3'
    description = _('This profile is intended for the Hanlin V3 and its clones.')

    # Screen size is a best guess
    screen_size = (584, 754)
    dpi = 168.451
    fbase = 16
    font_sizes = (12, 12, 14, 16, 18, 20, 22, 24)


class HanlinV5Output(HanlinV3Output):
    name = 'Hanlin V5'
    short_name = 'hanlinv5'
    description = _('This profile is intended for the Hanlin V5 and its clones.')

    dpi = 200


class CybookG3Output(OutputProfile):
    name = 'Cybook G3'
    short_name = 'cybookg3'
    description = _('This profile is intended for the Cybook G3.')

    # Screen size is a best guess
    screen_size = (600, 800)
    comic_screen_size = (600, 757)
    dpi = 168.451
    fbase = 16
    font_sizes = (12, 12, 14, 16, 18, 20, 22, 24)


class CybookOpusOutput(SonyReaderOutput):
    author = 'John Schember'
    name = 'Cybook Opus'
    short_name = 'cybook_opus'
    description = _('This profile is intended for the Cybook Opus.')

    # Screen size is a best guess
    dpi = 200
    fbase = 16
    font_sizes = (12, 12, 14, 16, 18, 20, 22, 24)

    epub_periodical_format = None


class KindleOutput(OutputProfile):
    name = 'Kindle'
    short_name = 'kindle'
    description = _('This profile is intended for the Amazon Kindle.')

    # Screen size is a best guess
    screen_size = (525, 640)
    dpi = 168.451
    fbase = 16
    font_sizes = (12, 12, 14, 16, 18, 20, 22, 24)
    supports_mobi_indexing = True
    periodical_date_in_title = False

    empty_ratings_char = '☆'
    ratings_char = '★'

    mobi_ems_per_blockquote = 2.0


class KindleDXOutput(OutputProfile):
    name = 'Kindle DX'
    short_name = 'kindle_dx'
    description = _('This profile is intended for the Amazon Kindle DX.')

    # Screen size is a best guess
    screen_size = (744, 1022)
    dpi = 150.0
    comic_screen_size = (771, 1116)
    # comic_screen_size         = (741, 1022)
    supports_mobi_indexing = True
    periodical_date_in_title = False
    empty_ratings_char = '☆'
    ratings_char = '★'
    mobi_ems_per_blockquote = 2.0


class KindlePaperWhiteOutput(KindleOutput):
    name = 'Kindle PaperWhite'
    short_name = 'kindle_pw'
    description = _('This profile is intended for the Amazon Kindle Paperwhite 1 and 2')

    # Screen size is a best guess
    screen_size = (658, 940)
    dpi = 212.0
    comic_screen_size = screen_size


class KindleVoyageOutput(KindleOutput):
    name = 'Kindle Voyage'
    short_name = 'kindle_voyage'
    description = _('This profile is intended for the Amazon Kindle Voyage')

    # Screen size is currently just the spec size, actual renderable area will
    # depend on someone with the device doing tests.
    screen_size = (1080, 1430)
    dpi = 300.0
    comic_screen_size = screen_size


class KindlePaperWhite3Output(KindleVoyageOutput):
    name = 'Kindle PaperWhite 3'
    short_name = 'kindle_pw3'
    description = _('This profile is intended for the Amazon Kindle Paperwhite 3 and above')
    # Screen size is currently just the spec size, actual renderable area will
    # depend on someone with the device doing tests.
    screen_size = (1072, 1430)
    dpi = 300.0
    comic_screen_size = screen_size


class KindleOasisOutput(KindlePaperWhite3Output):
    name = 'Kindle Oasis'
    short_name = 'kindle_oasis'
    description = _('This profile is intended for the Amazon Kindle Oasis 2017, Paperwhite 2021 and above')
    # Screen size is currently just the spec size, actual renderable area will
    # depend on someone with the device doing tests.
    screen_size = (1264, 1680)
    dpi = 300.0
    comic_screen_size = screen_size


class KindleScribeOutput(KindlePaperWhite3Output):
    name = 'Kindle Scribe'
    short_name = 'kindle_scribe'
    description = _('This profile is intended for the Amazon Kindle Scribe 2022 and above')
    # Screen size is currently just the spec size, actual renderable area will
    # depend on someone with the device doing tests.
    screen_size = (1860, 2480)
    dpi = 300.0
    comic_screen_size = screen_size


class KindleFireOutput(KindleDXOutput):
    name = 'Kindle Fire'
    short_name = 'kindle_fire'
    description = _('This profile is intended for the Amazon Kindle Fire.')

    screen_size = (570, 1016)
    dpi = 169.0
    comic_screen_size = (570, 1016)


class IlliadOutput(OutputProfile):
    name = 'Illiad'
    short_name = 'illiad'
    description = _('This profile is intended for the Irex Illiad.')

    screen_size = (760, 925)
    comic_screen_size = (760, 925)
    dpi = 160.0
    fbase = 12
    font_sizes = (7.5, 9, 10, 12, 15.5, 20, 22, 24)


class IRexDR1000Output(OutputProfile):
    author = 'John Schember'
    name = 'IRex Digital Reader 1000'
    short_name = 'irexdr1000'
    description = _('This profile is intended for the IRex Digital Reader 1000.')

    # Screen size is a best guess
    screen_size = (1024, 1280)
    comic_screen_size = (996, 1241)
    dpi = 160
    fbase = 16
    font_sizes = (12, 14, 16, 18, 20, 22, 24)


class IRexDR800Output(OutputProfile):
    author = 'Eric Cronin'
    name = 'IRex Digital Reader 800'
    short_name = 'irexdr800'
    description = _('This profile is intended for the IRex Digital Reader 800.')

    # Screen size is a best guess
    screen_size = (768, 1024)
    comic_screen_size = (768, 1024)
    dpi = 160
    fbase = 16
    font_sizes = (12, 14, 16, 18, 20, 22, 24)


class NookOutput(OutputProfile):
    author = 'John Schember'
    name = 'Nook'
    short_name = 'nook'
    description = _('This profile is intended for the B&N Nook.')

    # Screen size is a best guess
    screen_size = (600, 730)
    comic_screen_size = (584, 730)
    dpi = 167
    fbase = 16
    font_sizes = (12, 12, 14, 16, 18, 20, 22, 24)


class NookColorOutput(NookOutput):
    name = 'Nook Color'
    short_name = 'nook_color'
    description = _('This profile is intended for the B&N Nook Color.')

    screen_size = (600, 900)
    comic_screen_size = (594, 900)
    dpi = 169


class PocketBook900Output(OutputProfile):
    author = 'Chris Lockfort'
    name = 'PocketBook Pro 900'
    short_name = 'pocketbook_900'
    description = _('This profile is intended for the PocketBook Pro 900 series of devices.')

    screen_size = (810, 1180)
    dpi = 150.0
    comic_screen_size = screen_size


class PocketBookPro912Output(OutputProfile):
    author = 'Daniele Pizzolli'
    name = 'PocketBook Pro 912'
    short_name = 'pocketbook_pro_912'
    description = _('This profile is intended for the PocketBook Pro 912 series of devices.')

    # According to http://download.pocketbook-int.com/user-guides/E_Ink/912/User_Guide_PocketBook_912(EN).pdf
    screen_size = (825, 1200)
    dpi = 155.0
    comic_screen_size = screen_size


class PocketBookLuxOutput(OutputProfile):
    author = 'William Ouwehand'
    name = 'PocketBook Lux (1-5) and Basic 4'
    short_name = 'pocketbook_lux'
    description = _('This profile is intended for the PocketBook Lux (1-5) and Basic 4 series of devices.')

    screen_size = (758, 1024)
    dpi = 212.0
    comic_screen_size = screen_size


class PocketBookHDOutput(OutputProfile):
    author = 'William Ouwehand'
    name = 'PocketBook PocketBook HD Touch (1-3)'
    short_name = 'pocketbook_hd'
    description = _('This profile is intended for the PocketBook HD Touch (1-3) series of devices.')

    screen_size = (1072, 1448)
    dpi = 300.0
    comic_screen_size = screen_size


class PocketBookInkpad3Output(OutputProfile):
    author = 'William Ouwehand'
    name = 'PocketBook Inkpad 3 (Pro) and X'
    short_name = 'pocketbook_inkpad3'
    description = _('This profile is intended for the PocketBook Inkpad 3 and X series of devices.')

    screen_size = (1404, 1872)
    dpi = 227.0
    comic_screen_size = screen_size


output_profiles = [
    OutputProfile,
    SonyReaderOutput,
    SonyReader300Output,
    SonyReader900Output,
    SonyReaderT3Output,
    MSReaderOutput,
    MobipocketOutput,
    HanlinV3Output,
    HanlinV5Output,
    CybookG3Output,
    CybookOpusOutput,
    KindleOutput,
    iPadOutput,
    iPad3Output,
    KoboReaderOutput,
    TabletOutput,
    SamsungGalaxy,
    SonyReaderLandscapeOutput,
    KindleDXOutput,
    IlliadOutput,
    NookHD,
    IRexDR1000Output,
    IRexDR800Output,
    JetBook5Output,
    NookOutput,
    NookColorOutput,
    PocketBook900Output,
    PocketBookPro912Output,
    PocketBookLuxOutput,
    PocketBookHDOutput,
    PocketBookInkpad3Output,
    GenericEink,
    GenericEinkLarge,
    GenericEinkHD,
    KindleFireOutput,
    KindlePaperWhiteOutput,
    KindleVoyageOutput,
    KindlePaperWhite3Output,
    KindleOasisOutput,
    KindleScribeOutput,
]

output_profiles.sort(key=lambda x: x.name.lower())



#!/usr/bin/env python
# License: GPLv3 Copyright: 2008, Kovid Goyal kovid@kovidgoyal.net

from qt.core import QAbstractItemView, QDialog, Qt, QTableWidgetItem

from calibre import strftime
from calibre.ebooks.metadata import authors_to_sort_string, authors_to_string, title_sort
from calibre.gui2.dialogs.delete_matching_from_device_ui import Ui_DeleteMatchingFromDeviceDialog
from calibre.utils.date import UNDEFINED_DATE
from calibre.utils.localization import _


class tableItem(QTableWidgetItem):
    def __init__(self, text):
        QTableWidgetItem.__init__(self, text)
        self.setFlags(Qt.ItemFlag.ItemIsEnabled)
        self.sort = text.lower()

    def __ge__(self, other):
        return self.sort >= other.sort

    def __lt__(self, other):
        return self.sort < other.sort


class centeredTableItem(tableItem):
    def __init__(self, text):
        tableItem.__init__(self, text)
        self.setTextAlignment(Qt.AlignmentFlag.AlignCenter)


class titleTableItem(tableItem):
    def __init__(self, text):
        tableItem.__init__(self, text)
        self.sort = title_sort(text.lower())


class authorTableItem(tableItem):
    def __init__(self, book):
        tableItem.__init__(self, authors_to_string(book.authors))
        if book.author_sort is not None:
            self.sort = book.author_sort.lower()
        else:
            self.sort = authors_to_sort_string(book.authors).lower()


class dateTableItem(tableItem):
    def __init__(self, date):
        if date is not None:
            tableItem.__init__(self, strftime('%x', date))
            self.sort = date
        else:
            tableItem.__init__(self, '')
            self.sort = UNDEFINED_DATE


class DeleteMatchingFromDeviceDialog(QDialog, Ui_DeleteMatchingFromDeviceDialog):
    def __init__(self, parent, items):
        QDialog.__init__(self, parent)
        Ui_DeleteMatchingFromDeviceDialog.__init__(self)
        self.setupUi(self)

        self.explanation.setText('<p>' + _('All checked books will be <b>permanently deleted</b> from your device. Please verify the list.') + '</p>')
        self.buttonBox.accepted.connect(self.accepted)
        self.buttonBox.rejected.connect(self.rejected)
        self.table.cellClicked.connect(self.cell_clicked)
        self.table.setSelectionMode(QAbstractItemView.SelectionMode.NoSelection)
        self.table.setColumnCount(7)
        self.table.setHorizontalHeaderLabels([
            '',
            _('Location'),
            _('Title'),
            _('Author'),
            _('Date'),
            _('Format'),
            _('Path'),
        ])
        rows = 0
        for card in items:
            rows += len(items[card][1])
        self.table.setRowCount(rows)
        row = 0
        for card in items:
            model, books = items[card]
            for id, book in books:
                item = QTableWidgetItem()
                item.setFlags(Qt.ItemFlag.ItemIsUserCheckable | Qt.ItemFlag.ItemIsEnabled)
                item.setCheckState(Qt.CheckState.Checked)
                item.setData(Qt.ItemDataRole.UserRole, (model, id, book.path))
                self.table.setItem(row, 0, item)
                self.table.setItem(row, 1, tableItem(card))
                self.table.setItem(row, 2, titleTableItem(book.title))
                self.table.setItem(row, 3, authorTableItem(book))
                self.table.setItem(row, 4, dateTableItem(book.datetime))
                self.table.setItem(row, 5, centeredTableItem(book.path.rpartition('.')[2]))
                self.table.setItem(row, 6, tableItem(book.path))
                row += 1
        self.table.setCurrentCell(0, 1)
        self.table.resizeColumnsToContents()
        self.table.setSortingEnabled(True)
        self.table.sortByColumn(2, Qt.SortOrder.AscendingOrder)
        self.table.setCurrentCell(0, 1)

    def cell_clicked(self, row, col):
        if col == 0:
            self.table.setCurrentCell(row, 1)

    def accepted(self):
        self.result_val = []
        for row in range(self.table.rowCount()):
            item_0 = self.table.item(row, 0)
            assert item_0 is not None
            if item_0.checkState() == Qt.CheckState.Unchecked:
                continue
            model, id, path = item_0.data(Qt.ItemDataRole.UserRole)
            path = str(path)
            self.result_val.append((model, id, path))



# License: GPLv3 Copyright: 2010, Li Fanxi <lifanxi@freemindworld.com>

"""Read meta information from SNB files"""

import io
import os

from calibre.ebooks.metadata import MetaInformation
from calibre.ebooks.snb.snbfile import SNBFile
from calibre.utils.localization import _
from calibre.utils.xml_parse import safe_xml_fromstring


def get_metadata(stream, extract_cover=True):
    """Return metadata as a L{MetaInfo} object"""
    mi = MetaInformation(_('Unknown'), [_('Unknown')])
    snbFile = SNBFile()

    try:
        if not hasattr(stream, 'write'):
            snbFile.Parse(io.BytesIO(stream), True)
        else:
            stream.seek(0)
            snbFile.Parse(stream, True)

        meta = snbFile.GetFileStream('snbf/book.snbf')

        if meta is not None:
            meta = safe_xml_fromstring(meta)
            _name = meta.find('.//head/name')
            assert _name is not None
            mi.title = _name.text
            _author = meta.find('.//head/author')
            assert _author is not None
            mi.authors = [_author.text]
            _language = meta.find('.//head/language')
            assert _language is not None
            lang_text = _language.text
            assert lang_text is not None
            mi.language = lang_text.lower().replace('_', '-')
            _publisher = meta.find('.//head/publisher')
            assert _publisher is not None
            mi.publisher = _publisher.text

            if extract_cover:
                cover = meta.find('.//head/cover')
                if cover is not None and cover.text is not None:
                    root, ext = os.path.splitext(cover.text)
                    if ext == '.jpeg':
                        ext = '.jpg'
                    mi.cover_data = (ext[-3:], snbFile.GetFileStream('snbc/images/' + cover.text))

    except Exception:
        import traceback

        traceback.print_exc()

    return mi
